import collections
import functools
import inspect
import logging
import re

from .fetchable import *
from .compose import compose
from .currying import curry
from .misc import *

_pyflakes_ignore = (compose, curry, Fetchable, FetchMarshall,
                    GETFetcher, PaginationProcessor)
log = logging.getLogger(__name__)



def makelist(value):
    if not isinstance(value, list):
        if isinstance(value, collections.Iterable):
            return list(value)
        else:
            return [value]
    return value


def makelist_decorator(function):
    @functools.wraps(function)
    def wrapped(arg):
        return function(makelist(arg))

    return wrapped


class cached_property(object):

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = self.func(obj)
        setattr(obj, self.func.__name__, value)
        return value

    def bust_self(self, obj):
        if self.func.__name__ in obj.__dict__:
            delattr(obj, self.func.__name__)

    @classmethod
    def bust_caches(cls, obj):
        for name, _ in cls.get_cached_properties(obj):
            if name in obj.__dict__:
                delattr(obj, name)

    @classmethod
    def get_cached_properties(cls, obj):
        return inspect.getmembers(type(obj), lambda x: isinstance(x, cls))


class CallableMap(object):

    def __init__(self, func_value_pairs=()):
        if isinstance(func_value_pairs, dict):
            func_value_pairs = func_value_pairs.items()
        self.func_value_pairs = list(func_value_pairs)

    def __getitem__(self, item):
        for func, value in self.func_value_pairs:
            if func(item):
                return value
        raise KeyError('{0} did not match any of this objects functions.'.format(item))

    def __setitem__(self, function, value):
        self.add(func, value)

    def add(func, value):
        self.func_value_pairs.append((func, value))


class REMap(object):

    NO_DEFAULT = object()

    @classmethod
    def from_string_pairs(cls, string_value_pairs, **kwargs):
        return cls(re_value_pairs=[(re.compile(s), v)
                                   for s, v in string_value_pairs],
                   **kwargs)

    def __init__(self, re_value_pairs=(), default=NO_DEFAULT):
        self.default = default
        if isinstance(re_value_pairs, dict):
            re_value_pairs = re_value_pairs.items()
        self.re_value_pairs = list(re_value_pairs)

    def __getitem__(self, item):
        if item is None:
            if self.default is self.NO_DEFAULT:
                raise KeyError("None does not match any expression")
            else:
                return self.default
        for matcher, value in self.re_value_pairs:
            if matcher.search(item):
                return value
        if self.default is not self.NO_DEFAULT:
            if item is not None and len(item) > 1:
                log.warning("Returning default from REMAP for {0}.".format(
                    repr(item)
                ))
            return self.default
        else:
            raise KeyError('{0} did not match any of this objects regular'
                           'expressions.'.format(repr(item)))

    def __setitem__(self, re, value):
        self.add(re, value)

    def add(re, value):
        self.re_value_pairs.append((re, value))

    @property
    def pattern_to_value(self):
        return {expression.pattern: value
                for expression, value in self.re_value_pairs}




import okcupyd


u = okcupyd.User()
p = u.quickmatch()

# How to use other users questions:
for q in p.questions:
    if 'marijuana' in q.text.lower():
        print(q.text)
        print(q.their_answer)
        if q.their_note and len(q.their_not):
            print(q.their_note)
        print('')



import logging

from .question import Questions


log = logging.getLogger(__name__)


class Copy(object):

    copy_methods = ['questions', 'photos', 'essays', 'looking_for', 'details']

    def __init__(self, source_user, dest_user):
        self.source_user = source_user
        self.dest_user = dest_user

    def questions(self):
        dest_questions = self.dest_user.questions
        for key, importance in Questions.importance_name_to_number.items():
            questions = getattr(self.source_user.questions, key)
            for question in questions:
                log.debug(
                    dest_questions.respond_from_user_question(
                        question,
                        importance
                    ).content
                )

    def photos(self):
        source_user_profile = self.dest_user.get_profile(
            self.source_user.profile.username
        )
        # Reverse because pictures appear in inverse chronological order.
        return [self.dest_user.photo.upload_and_confirm(info)
                for info in reversed(source_user_profile.photo_infos)]

    def essays(self):
        for essay_name in self.dest_user.profile.essays.essay_names:
            setattr(self.dest_user.profile.essays, essay_name,
                    getattr(self.source_user.profile.essays, essay_name))

    def looking_for(self):
        looking_for = self.source_user.profile.looking_for
        return self.dest_user.profile.looking_for.update(
            gentation=looking_for.gentation,
            single=looking_for.single,
            near_me=looking_for.near_me,
            kinds=looking_for.kinds,
            ages=looking_for.ages
        )

    def details(self):
        return self.dest_user.profile.details.convert_and_update(
            self.source_user.profile.details.as_dict
        )

    def all(self):
        for method_name in self.copy_methods:
            getattr(self, method_name)()



from okcupyd.db import txn, model
from okcupyd.db.adapters import ThreadAdapter


def test_thread_adapter_create_and_update(T):
    initiator = 'first_initiator'
    respondent = 'respondent_one'

    message_thread = T.build_mock.thread(initiator=initiator,
                                         respondent=respondent)

    thread_model = ThreadAdapter(message_thread).get_thread()
    T.ensure.thread_model_resembles_okcupyd_thread(
        thread_model, message_thread
    )

    # Ensure that the operation is idempotent
    other_thread_model = ThreadAdapter(message_thread).get_thread()
    T.ensure.thread_model_resembles_okcupyd_thread(
        thread_model, message_thread
    )
    assert other_thread_model.id == thread_model.id

    # Add messages and ensure that they are picked up.
    message_thread.messages.append(
        T.build_mock.message(sender=message_thread.initiator.username,
                             recipient=message_thread.respondent.username,
                             content='other')
    )

    ThreadAdapter(message_thread).get_thread()
    with txn() as session:
        loaded_thread_model = model.MessageThread.find_no_txn(session,
                                                              thread_model.id)
        assert loaded_thread_model.id == thread_model.id
        T.ensure.thread_model_resembles_okcupyd_thread(
            loaded_thread_model, message_thread
        )
        for i, message in enumerate(loaded_thread_model.messages):
            assert message.thread_index == i

    assert other_thread_model.id == thread_model.id

    second_thread = T.build_mock.thread(initiator=initiator,
                                        respondent='new_respondent')

    second_thread_model = ThreadAdapter(second_thread).get_thread()
    assert second_thread_model.initiator.id == thread_model.initiator.id
    assert second_thread_model.respondent != thread_model.initiator

    T.ensure.thread_model_resembles_okcupyd_thread(second_thread_model,
                                                   second_thread)



from okcupyd.search import SearchFetchable, search
from okcupyd.profile import Profile
from . import util


@util.use_cassette(cassette_name='search_age_filter')
def test_age_filter():
    age = 22
    search_fetchable = SearchFetchable(gentation='everybody',
                                       age_min=age, age_max=age)

    profile = next(iter(search_fetchable))
    assert profile.age == age


@util.use_cassette(cassette_name='search_count')
def test_count_variable(request):
    profiles = search(gentation='everybody', count=14)
    assert len(profiles) == 14

    for profile in profiles:
        profile.username
        profile.age
        profile.location
        profile.match_percentage
        profile.enemy_percentage
        profile.id
        profile.rating
        profile.contacted


@util.use_cassette(cassette_name='search_location_filter')
def test_location_filter():
    location = 'Portland, OR'
    search_fetchable = SearchFetchable(location=location, radius=25)
    profile = search_fetchable[0]
    assert profile.location == 'Portland, OR'


@util.use_cassette(cassette_name='search_function')
def test_search_function():
    profile, = search(count=1)
    assert isinstance(profile, Profile)

    profile.username
    profile.age
    profile.location
    profile.match_percentage
    profile.enemy_percentage
    profile.id
    profile.rating
    profile.contacted


@util.use_cassette
def test_search_fetchable_iter():
    search_fetchable = SearchFetchable(gentation='everybody',
                                       religion='buddhist', age_min=25, age_max=25,
                                       location='new york, ny', keywords='bicycle')
    for count, profile in enumerate(search_fetchable):
        assert isinstance(profile, Profile)
        if count > 30:
            break



import pytest

from okcupyd import settings
from okcupyd.session import Session
from okcupyd.errors import AuthenticationError
from . import util


@util.use_cassette(cassette_name='session_success')
def test_session_success():
    Session.login()


@util.use_cassette(cassette_name='session_failure')
def test_session_auth_failure():
    with pytest.raises(AuthenticationError):
        Session.login(password='wrong_password')



import logging

import requests

from .errors import AuthenticationError
from . import settings
from . import util
from . import profile


log = logging.getLogger(__name__)


class Session(requests.Session):
    """A :ref:`requests.Session` with convenience methods for interacting with
    okcupid.com"""

    default_login_headers = {
        'user-agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_4) '
                       'AppleWebKit/537.36 (KHTML, like Gecko) '
                       'Chrome/37.0.2062.94 '
                       'Safari/537.36')
    }

    @classmethod
    def login(cls, username=None, password=None):
        """
        :param username: The username to log in with.
        :type username: str
        :param password: The password to log in with.
        :type password: str
        Get a session that has authenticated with okcupid.com.
        If no username or password is supplied, the
        """
        # settings.USERNAME and settings.PASSWORD should not be made
        # the defaults to their respective arguments because doing so
        # would prevent this function from picking up any changes made
        # to those values after import time.
        username = username or settings.USERNAME
        password = password or settings.PASSWORD
        session = cls()
        credentials = {
            'username': username,
            'password': password,
            'okc_api': 1
        }
        login_response = session.okc_post('login',
                                          data=credentials,
                                          headers=cls.default_login_headers,
                                          secure=True)
        log_in_name = login_response.json()['screenname']
        if log_in_name is None:
            raise AuthenticationError('Could not log in as {0}'.format(username))
        if log_in_name.lower() != username.lower():
            log.warning('Expected to log in as {0} but '
                        'got {1}'.format(username, log_in_name))
        log.debug(login_response.content.decode('utf8'))
        session.log_in_name = log_in_name
        session.headers.update(cls.default_login_headers)
        return session

    def get(self, *args, **kwargs):
        response = super(Session, self).get(*args, **kwargs)
        response.raise_for_status()
        return response

    def post(self, *args, **kwargs):
        log.debug(kwargs)
        response = super(Session, self).post(*args, **kwargs)
        response.raise_for_status()
        log.debug(response.content)
        return response

    def okc_get(self, path, secure=None, **kwargs):
        """Perform an HTTP GET to the www.okcupid.com at the provided path."""
        response = self.get(self.build_path(path, secure), **kwargs)
        return response

    def okc_post(self, path, secure=None, **kwargs):
        """Perform an HTTP POST to the www.okcupid.com at the provided path."""
        return self.post(self.build_path(path, secure), **kwargs)

    def build_path(self, path, secure=None):
        if secure is None:
            secure = ('secure_login' in self.cookies and
                      int(self.cookies['secure_login']) != 0)
        return u'{0}://{1}/{2}'.format('https' if secure else 'http',
                                       util.DOMAIN, path)

    def get_profile(self, username):
        """Get the profile associated with the supplied username
        :param username: The username of the profile to retrieve."""
        return profile.Profile(self, username)

    def get_current_user_profile(self):
        """Get the `okcupyd.profile.Profile`  associated with the supplied
        username.
        :param username: The username of the profile to retrieve.
        """
        return self.get_profile(self.log_in_name)



"""Give a 5-star rating to 3 lucky 24-year-old straight/gay/bi women/men."""
import okcupyd

u = okcupyd.User()
profiles = u.search(location='minneapolis, mn', keywords='arrested development',
                    age_min=24, age_max=24, number=20)

for profile in profiles[:3]:
    profile.rate(5)



import datetime

from . import util
from okcupyd import looking_for
from okcupyd import profile
from okcupyd import session
from okcupyd import User
from okcupyd.util import cached_property


@util.use_cassette
def test_profile_essays():
    user_profile = profile.Profile(session.Session.login(), 'FriedLiverAttack')
    assert user_profile.essays.self_summary


@util.use_cassette
def test_rate_profile():
    profile = User().quickmatch()
    profile.rate(5)
    profile.refresh()
    assert profile.rating == 5


@util.use_cassette
def test_rate_profile_1_stars():
    profile = User().quickmatch()
    profile.rate(1)
    assert User().get_profile(profile.username).rating == profile.rating
    assert profile.rating == 1


@util.use_cassette
def test_profile_properties():
    profile = User().quickmatch()
    assert 0 <= profile.match_percentage <= 100
    assert 0 <= profile.enemy_percentage <= 100
    assert profile.responds
    assert profile.contacted == False # We want it to be false, not falsy


@util.use_cassette
def test_profile_attractiveness():
    # TODO: get rid of circular import so you can do this.
    profile = User().quickmatch()
    assert 1 < profile.attractiveness < 10001


@util.use_cassette
def test_profile_contacted():
    profile = User().quickmatch()
    profile.message('test')
    assert bool(profile.contacted)
    assert isinstance(profile.contacted, datetime.datetime)


@util.use_cassette
def test_profile_on_inbox_correspondent():
    profile = User().outbox[-1].correspondent_profile
    assert bool(profile.contacted)
    assert isinstance(profile.contacted, datetime.datetime)
    for prop_name, _ in cached_property.get_cached_properties(profile):
        getattr(profile, prop_name)


@util.use_cassette
def test_looking_for_on_user_profile():
    profile = User().profile
    min, max = profile.looking_for.ages
    assert isinstance(profile.looking_for.single, bool)
    assert isinstance(profile.looking_for.near_me, bool)
    assert isinstance(profile.looking_for.kinds, list)


@util.use_cassette
def test_looking_for_write_on_user_profile(vcr_live_sleep):
    profile = User().profile
    new_single = not profile.looking_for.single
    new_ages_min, new_ages_max = profile.looking_for.ages
    new_ages_min += 1
    new_ages_max += 1
    new_near_me = not profile.looking_for.near_me
    new_kinds = (['long-term dating', 'casual sex']
                 if 'short-term dating' in profile.looking_for.kinds
                 else ['short-term dating', 'casual sex'])
    new_gentation = ('bi girls only' if 'everybody' in profile.looking_for.gentation
                     else 'everybody')

    sleep_time = 4
    profile.looking_for.ages = new_ages_min, new_ages_max
    vcr_live_sleep(sleep_time)
    profile.looking_for.near_me = new_near_me
    vcr_live_sleep(sleep_time)
    profile.looking_for.kinds = new_kinds
    vcr_live_sleep(sleep_time)
    profile.looking_for.gentation = new_gentation
    vcr_live_sleep(sleep_time)
    profile.looking_for.single = new_single

    vcr_live_sleep(sleep_time)
    new_profile = profile._session.get_profile(profile.username)
    assert new_profile.looking_for.single == new_single
    assert new_profile.looking_for.near_me == new_near_me
    assert new_profile.looking_for.ages == looking_for.LookingFor.Ages(new_ages_min,
                                                                       new_ages_max)
    assert set(new_profile.looking_for.kinds) == set(new_kinds)
    assert new_profile.looking_for.gentation == new_gentation


    assert profile.looking_for.single == new_single
    assert profile.looking_for.near_me == new_near_me
    assert profile.looking_for.ages == looking_for.LookingFor.Ages(new_ages_min,
                                                                   new_ages_max)
    assert set(profile.looking_for.kinds) == set(new_kinds)
    assert profile.looking_for.gentation == new_gentation


@util.use_cassette
def test_details():
    pass



import inspect


class curry(object):

    @staticmethod
    def arity_evaluation_checker(function):
        is_class = inspect.isclass(function)
        if is_class:
            function = function.__init__
        function_info = inspect.getargspec(function)
        function_args = function_info.args
        if is_class:
            # This is to handle the fact that self will get passed in
            # automatically.
            function_args = function_args[1:]
        def evaluation_checker(*args, **kwargs):
            kwarg_keys = set(kwargs.keys())
            if function_info.keywords == None:
                acceptable_kwargs = function_args[len(args):]
                # Make sure that we didn't get an argument we can't handle.
                if not kwarg_keys.issubset(acceptable_kwargs):
                    TypeError("Unrecognized Arguments: {0}".format(
                        [key for key in kwarg_keys
                         if key not in acceptable_kwargs]
                    ))
            needed_args = function_args[len(args):]
            if function_info.defaults:
                needed_args = needed_args[:-len(function_info.defaults)]
            return not needed_args or kwarg_keys.issuperset(needed_args)
        return evaluation_checker

    @staticmethod
    def count_evaluation_checker(count):
        def function(*args, **kwargs):
            return len(args) >= count
        return function

    def __init__(self, function, evaluation_checker=None, args=(), kwargs=None):
        self.function = function
        self.evaluation_checker = (evaluation_checker or
                                   self.arity_evaluation_checker(function))
        self.args = args
        self.kwargs = kwargs or {}
        self.__name__ = function.__name__

    def __call__(self, *args, **kwargs):
        new_args = self.args + args
        new_kwargs = self.kwargs.copy()
        new_kwargs.update(kwargs)
        if self.evaluation_checker(*new_args, **new_kwargs):
            return self.function(*new_args, **new_kwargs)
        else:
            return type(self)(self.function, self.evaluation_checker,
                              new_args, new_kwargs)

    def __get__(self, obj, obj_type):
        bound = type(self)(self.function, self.evaluation_checker,
                           args=self.args + (obj,), kwargs=self.kwargs)
        setattr(obj, self.function.__name__, bound)
        return bound

    def __repr__(self):
        return '<{0}.{1} of {2}>'.format(__name__, type(self).__name__, 
                                  repr(self.function))


curry = curry(curry)



from okcupyd.db import user


def test_have_messaged_before(T):
    message_thread_model = T.factory.message_thread()
    assert user.have_messaged_by_username(
        message_thread_model.initiator.handle,
        message_thread_model.respondent.handle
    )
    assert user.have_messaged_by_username(
        message_thread_model.respondent.handle,
        message_thread_model.initiator.handle
    )

    assert not user.have_messaged_by_username('a', 'b')
    assert not user.have_messaged_by_username(
        message_thread_model.respondent.handle, 'a'
    )

    T.factory.user('b')
    assert not user.have_messaged_by_username(
        'b', message_thread_model.initiator.handle
    )



"""Read the first message in your inbox and reply in two different ways."""

import okcupyd

u = okcupyd.User()
thread = u.outbox[0]
print(thread.messages)
u.message(thread.correspondent, 'Thank you for that highly informative message!')
thread.reply('This is another way to send a message')



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint

from okcupyd.db import OKCBase


class Message(OKCBase):

    __tablename__ = "message"

    __table_args__ = (
        UniqueConstraint('message_thread_id', 'thread_index'),
    )

    message_thread_id = Column(Integer, ForeignKey("message_thread.id"),
                               nullable=False)

    sender_id = Column(Integer, ForeignKey("user.id"))
    sender = relationship("User", foreign_keys=[sender_id])

    recipient_id = Column(Integer, ForeignKey("user.id"))
    recipient = relationship("User", foreign_keys=[recipient_id])

    text = Column(String, nullable=False)
    thread_index = Column(Integer, nullable=False)



from lxml import etree

from okcupyd import xpath


def test_selected_attribute():
    node = xpath.XPathNode(element='element', selected_attribute='value')
    assert node.xpath == '//element/@value'

    tree = etree.XML("<top><container><element value='1'>"
                     "</element><element value='2'></element></container></top>")
    builder = xpath.xpb.container.element.select_attribute_('value')

    assert builder.xpath == './/container//element/@value'
    assert builder.apply_(tree) == ['1', '2']

    assert xpath.xpb.element.select_attribute_('value', elem=tree) == ['1', '2']


def test_text_for_many():
    tree = etree.XML("<top><container>"
                     "<element value='1'>one</element>"
                     "<element value='2'>two</element>"
                     "</container></top>")

    result = xpath.xpb.container.element.text_.apply_(tree)
    assert set(result) == set(['one', 'two'])


def test_attribute_contains():
    tree = etree.XML("<top><elem a='complete'></elem></top>")
    assert xpath.xpb.elem.attribute_contains('a', 'complet').apply_(tree) != []



import logging

from invoke import Collection, task
import IPython

from . import copy
from . import db as db_collection
from okcupyd import user, db, util


log = logging.getLogger(__name__)


ns = Collection()
ns.add_collection(copy)
ns.add_collection(db_collection)

login = task(util.get_credentials)
ns.add_task(login, 'login')


@ns.add_task
@task(login, default=True)
def interactive():
    u = user.User()
    IPython.embed()


@ns.add_task
@task(login)
def enable_logger(logger_name):
    util.enable_logger(logger_name)


@ns.add_task
@task(login, aliases=('c',))
def credentials(module_name):
    util.update_settings_with_module(module_name)


@ns.add_task
@task(login, aliases=('eal'))
def enable_all_loggers():
    for logger_name in ('okcupyd', 'requests', __name__):
        util.enable_logger(logger_name)
    db.Session.kw['bind'].echo = True





import collections
import functools

from .currying import curry

def _compose2(f, g):
    return lambda *args, **kwargs: f(g(*args, **kwargs))


@curry(evaluation_checker=curry.count_evaluation_checker(2))
def compose_with_joiner(joiner, *functions):
    return functools.reduce(joiner, functions)


compose_one_arg = compose_with_joiner(_compose2)


compose = compose_with_joiner(lambda f, g: _compose2(make_single_arity(f),
                                                force_args_return(g)))

def make_single_arity(function):
    @functools.wraps(function)
    def wrapped(args):
        return function(*args)
    return wrapped


def force_args_return(function):
    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        value = function(*args, **kwargs)
        if not isinstance(value, collections.Iterable):
            value = (value,)
        return value
    return wrapped


def tee(*functions):
    def wrapped(*args, **kwargs):
        return tuple(function(*args, **kwargs) for function in functions)
    return wrapped



from . import util
from okcupyd import User


@util.use_cassette
def test_message_thread_detect_deletion():
    message_thread = User().inbox[0]
    assert message_thread.with_deleted_user
    sent_thread = User().outbox[0]
    assert not sent_thread.with_deleted_user


@util.use_cassette
def test_initiator_and_respondent():
    user = User()
    their_profile = user.quickmatch()
    message_info = their_profile.message('initiated contact')
    while message_info.message_id is None:
        their_profile = user.quickmatch()
        message_info = their_profile.message('initiated contact')

    message_thread = user.outbox[0]

    assert message_thread.initiator.username == user.profile.username
    assert message_thread.respondent.username == their_profile.username


@util.use_cassette
def test_load_messages_on_message_thread():
    user = User()
    content = 'a message was sent'
    their_profile = user.quickmatch()
    message_info = their_profile.message(content)
    while message_info.message_id is None:
        their_profile = user.quickmatch()
        message_info = their_profile.message(content)

    assert len(user.outbox[0].messages) == 1
    assert user.outbox[0].messages[0].content == content



import datetime
import logging


from lxml import html
import simplejson

from . import details
from . import essay
from . import helpers
from . import looking_for
from . import util
from .question import QuestionFetcher
from .xpath import xpb


log = logging.getLogger(__name__)


class Profile(object):
    """Represent the profile of an okcupid user.

    Many of the attributes on this object are
    :class:`okcupyd.util.cached_property` instances which lazily load their
    values, and cache them once they have been accessed. This makes it so
    that this object avoids making unnecessary HTTP requests to retrieve the
    same piece of information twice.

    Because of this caching behavior, care must
    be taken to invalidate cached attributes on the object if an up to date view
    of the profile is needed. It is recommended that you call :meth:`refresh` to
    accomplish this, but it is also possible to use
    :meth:`okcupyd.util.cached_property.bust_self` to bust individual propeties
    if necessary.
    """

    def __init__(self, session, username):
        """
        :param session: A logged in :class:`okcupyd.session.Session`
        :param username: The username associated with the profile.
        """
        self._session = session
        self.username = username
        self.questions = self.question_fetchable()
        self.details = details.Details(self)

    def refresh(self, reload=False):
        """
        :param reload: Make the request to return a new profile tree. This will
                       result in the caching of the profile_tree attribute. The
                       new profile_tree will be returned.
        """
        util.cached_property.bust_caches(self)
        self.questions = self.question_fetchable()
        if reload:
            return self.profile_tree

    @property
    def is_logged_in_user(self):
        """Return True if this profile and the session it was created with
        belong to the same user and False otherwise."""
        return self._session.log_in_name.lower() == self.username.lower()

    @util.cached_property
    def _profile_response(self):
        return self._session.okc_get(
            u'profile/{0}'.format(self.username)
        ).content

    @util.cached_property
    def profile_tree(self):
        """Return a :class:`lxml.etree` created from the html of the profile
        page of the account associated with the username that this profile was
        insantiated with.
        """
        return html.fromstring(self._profile_response)

    def message_request_parameters(self, content, thread_id):
        return {
            'ajax': 1,
            'sendmsg': 1,
            'r1': self.username,
            'body': content,
            'threadid': thread_id,
            'authcode': self.authcode,
            'reply': 1 if thread_id else 0,
            'from_profile': 1
        }

    @util.cached_property
    def authcode(self):
        return helpers.get_authcode(self.profile_tree)

    @util.cached_property
    def photo_infos(self):
        """
        :returns: list of :class:`okcupyd.photo.Info` instances for each photo
        displayed on okcupid.
        """
        pics_request = self._session.okc_get(
            u'profile/{0}/photos#0'.format(self.username)
        )
        pics_tree = html.fromstring(pics_request.content.decode('utf8'))
        from . import photo
        return map(photo.Info.from_cdn_uri,
                   xpb.div(id='album_0').img.select_attribute_('src',
                                                               pics_tree))

    @util.cached_property
    def looking_for(self):
        """Return a :class:`okcupyd.looking_for.LookingFor` instance associated
        with this profile.
        """
        return looking_for.LookingFor(self)

    _rating_xpb = xpb.div(id='rating').ul(id='personality-rating').li.\
                  with_class('current-rating')

    @util.cached_property
    def rating(self):
        """Return the rating that the logged in user has given this user or
        0 if no rating has been given.
        """
        if self.is_logged_in_user: return 0
        rating_style = self._rating_xpb.select_attribute_('style').one_(
            self.profile_tree
        )
        width_percentage = int(''.join(c for c in rating_style if c.isdigit()))
        return width_percentage // 20

    _contacted_xpb = xpb.div(id='actions').div.with_classes('tooltip_text',
                                                               'hidden')

    @util.cached_property
    def contacted(self):
        """Return a boolean indicating whether the logged in user has contacted
        the owner of this profile.
        """
        try:
            contacted_span = self._contacted_xpb.span.with_class('fancydate').\
                             one_(self.profile_tree)
        except:
            return False
        else:
            timestamp = contacted_span.attrib['id'].split('_')[-1][:-2]
            return datetime.datetime.fromtimestamp(int(timestamp[:10]))

    @util.cached_property
    def responds(self):
        """Return the frequency with which the user associated with this profile
        responds to messages.
        """
        contacted_text = self._contacted_xpb.\
                         get_text_(self.profile_tree).lower()
        if 'contacted' not in contacted_text:
            return contacted_text.strip().replace('replies ', '')

    @util.cached_property
    def id(self):
        """The id that okcupid.com associates with this profile."""
        if self.is_logged_in_user: return self._current_user_id
        return int(self._rating_xpb.select_attribute_('id').
                   one_(self.profile_tree).split('-')[-2])

    @util.cached_property
    def _current_user_id(self):
        return int(helpers.get_id(self.profile_tree))

    @util.cached_property
    def essays(self):
        """An :class:`okcupyd.essay.Essays` instance that is associated with
        this profile.
        """
        return essay.Essays(self)

    @util.cached_property
    def age(self):
        """The age of the user associated with this profile."""
        return int(xpb.span(id='ajax_age').get_text_(self.profile_tree).strip())

    _percentages_and_ratings_xpb = xpb.div(id='percentages_and_ratings')

    @util.cached_property
    def match_percentage(self):
        """The match percentage of the logged in user and the user associated
        with this object.
        """
        return int(self._percentages_and_ratings_xpb.
                   div.with_class('match').
                   span.with_class('percent').
                   get_text_(self.profile_tree).strip('%'))

    @util.cached_property
    def enemy_percentage(self):
        """The enemy percentage of the logged in user and the user associated
        with this object.
        """
        return int(self._percentages_and_ratings_xpb.
                   div.with_class('enemy').
                   span.with_class('percent').
                   get_text_(self.profile_tree).strip('%'))

    @util.cached_property
    def location(self):
        """The location of the user associated with this profile."""
        return xpb.span(id='ajax_location').get_text_(self.profile_tree)

    @util.cached_property
    def gender(self):
        """The gender of the user associated with this profile."""
        return xpb.span(id='ajax_gender').get_text_(self.profile_tree)

    @util.cached_property
    def orientation(self):
        """The sexual orientation of the user associated with this profile."""
        return xpb.dd(id='ajax_orientation').\
            get_text_(self.profile_tree).strip()

    @util.curry
    def message(self, message, thread_id=None):
        """Message the user associated with this profile.
        :param message: The message to send to this user.
        :param thread_id: The id of the thread to respond to, if any.
        """
        return_value =  helpers.Messager(
            self._session
        ).send(self.username, message,
                       self.authcode, thread_id)
        self.refresh(reload=False)
        return return_value

    @util.cached_property
    def attractiveness(self):
        """The average attractiveness rating given to this profile by the
        okcupid.com community.
        """
        from .attractiveness_finder import AttractivenessFinder
        return AttractivenessFinder(self._session)(self.username)

    def rate(self, rating):
        """Rate this profile as the user that was logged in with the session
        that this object was instantiated with.
        :param rating: The rating to give this user.
        """
        parameters = {
            'voterid': self._current_user_id,
            'target_userid': self.id,
            'type': 'vote',
            'cf': 'profile2',
            'target_objectid': 0,
            'vote_type': 'personality',
            'score': rating,
        }
        response = self._session.okc_post('vote_handler',
                                          data=parameters)
        response_json = response.json()
        log_function = log.info if response_json.get('status', False) \
                       else log.error
        log_function(simplejson.dumps({'rate_response': response_json,
                                       'sent_parameters': parameters,
                                       'headers': dict(self._session.headers)}))
        self.refresh(reload=False)

    def question_fetchable(self, **kwargs):
        """Return a :class:`okcupyd.util.fetchable.Fetchable` instance that
        contains objects representing the answers that the user associated with
        this profile has given to okcupid.com match questions.
        """
        return util.Fetchable(QuestionFetcher(
            self._session, self.username,
            is_user=self.is_logged_in_user, **kwargs
        ))

    def authcode_get(self, path, **kwargs):
        """Perform an HTTP GET to okcupid.com using this profiles session
        where the authcode is automatically added as a query parameter.
        """
        kwargs.setdefault('params', {})['authcode'] = self.authcode
        return self._session.okc_get(path, **kwargs)

    def authcode_post(self, path, **kwargs):
        """Perform an HTTP POST to okcupid.com using this profiles session
        where the authcode is automatically added as a form item.
        """
        kwargs.setdefault('data', {})['authcode'] = self.authcode
        return self._session.okc_post(path, **kwargs)

    def __eq__(self, other):
        self.username.lower() == other.username.lower()

    def __repr__(self):
        return '{0}("{1}")'.format(type(self).__name__, self.username)



from sqlalchemy import Column, DateTime, ForeignKey, Integer, String

from okcupyd.db import Base, OKCBase


class Question(OKCBase):

    __tablename__ = 'question'

    text = Column(String, nullable=False)


class UserQuestionAnswerOption(Base):

    user_id = Column(Integer, ForeignKey("user.id"))



import inspect

from . import util
from . import magicnumbers


class Filters(object):
    """Registers kwargs that can be used to construct filters for submission
    in requests to okcupid.com
    """

    def __init__(self):
        self.builder_to_keys = {}
        self.builder_to_decider = {}
        self.keys = set()
        self._key_to_type = {}
        self._key_to_values = {}
        self._key_to_string = {}

    def build_documentation(self):
        return u'\n'.join(self.build_paramter_string(key)
                         for key in sorted(self.keys))

    def build_paramter_string(self, key):
        description_string = u''
        if key in self._key_to_string:
            description_string = u' {0}'.format(self._key_to_string[key])
        if key in self._key_to_values:
            description_string += u' Allowed values: {0}'.format(
                u', '.join([repr(value)
                           for value in self._key_to_values[key]])
            )
        parameter_string = u':param {0}:{1}'.format(
            key, description_string
        )
        if key in self._key_to_type:
            parameter_string += u'\n'
            the_type = self._key_to_type[key]
            parameter_string += u':type {0}: {1}'.format(
                key, the_type.__name__ if isinstance(the_type, type) else the_type
            )
        return parameter_string

    @staticmethod
    def any_decider(function, incoming, accepted_keys):
        return bool(set(incoming).intersection(accepted_keys))

    @staticmethod
    def all_decider(function, incoming, accepted_keys):
        return set(accepted_keys).issubset(set(incoming))

    @staticmethod
    def all_not_none_decider(function, incoming, accepted_keys):
        return all(incoming.get(key) is not None for key in accepted_keys)

    @staticmethod
    def any_not_none_decider(function, incoming, accepted_keys):
        return any(incoming.get(key) is not None for key in accepted_keys)

    def __repr__(self):
        return "{0}({1})".format(type(self).__name__,
                                 repr(self.builder_to_keys.keys()))

    def filters(self, **kwargs):
        builders = [builder
                    for builder, decider in self.builder_to_decider.items()
                    if decider(builder, kwargs, self.builder_to_keys[builder])]
        return [builder(*[kwargs.get(key)
                          for key in self.builder_to_keys[builder]])
                for builder in builders]

    def build(self, **kwargs):
        return {
            u'filter{0}'.format(filter_number): filter_string
            for filter_number, filter_string
            in enumerate(self.filters(**kwargs), 1)
        }

    @util.curry
    def register_filter_builder(self, function, keys=(), decider=None,
                                acceptable_values=None,
                                types=None,
                                descriptions=None):
        decider = decider or self.all_not_none_decider
        function_arguments = inspect.getargspec(function).args
        if keys:
            assert len(keys) == len(function_arguments)
        else:
            keys = function_arguments
        self.builder_to_keys[function] = keys
        self.builder_to_decider[function] = decider

        self.keys.update(keys)
        self._update_docs_dict(self._key_to_type, types, keys)
        self._update_docs_dict(self._key_to_string, descriptions, keys)
        self._update_docs_dict(self._key_to_values, acceptable_values,
                               keys, unpack_lists=False)

        return function

    def _update_docs_dict(self, docs_dict, incoming, keys, unpack_lists=True):
        if incoming:
            if not isinstance(incoming, dict):
                if isinstance(incoming, (tuple, list)) and unpack_lists:
                    assert len(keys) == len(incoming), (
                        "Got {0}, for keys: {1}".format(
                            incoming, keys
                        )
                    )
                    incoming = zip(keys, incoming)
                else:
                    incoming = {key: incoming for key in keys}
            docs_dict.update(incoming)


def gentation_filter(gentation):
    return u'0,{0}'.format(magicnumbers.gentation_to_number[gentation.lower()])


def age_filter(age_min=18, age_max=99):
    if age_min == None:
        age_min = 18
    return u'2,{0},{1}'.format(age_min, age_max)


def location_filter(radius):
    return u'3,{0}'.format(radius)



"""Sample script that downloads the profile pictures of all 5'6"-5'8" Buddhists
living in New York, NY who either do not smoke or are trying to quit.
"""
import okcupyd

session = okcupyd.Session.login()
u = okcupyd.User(session)
profiles = u.search(location='new york, ny', religion='buddhist',
                     height_min=66, height_max=68, gentation='everybody',
                     smokes=['no', 'trying to quit'])


for profile in profiles[:2]:
    for photo_info in profile.photo_infos:
        print(photo_info.jpg_uri)



import logging

from lxml import html
from requests import exceptions

from . import helpers
from . import util
from .xpath import XPathBuilder, xpb


log = logging.getLogger(__name__)

thread_element_xpath = XPathBuilder(relative=False).li.with_classes('thread',
                                                                    'message')
def ThreadFetcher(session, mailbox_number):
    return util.FetchMarshall(
        ThreadHTMLFetcher(session, mailbox_number),
        util.SimpleProcessor(
            session, lambda elem: MessageThread(session, elem), thread_element_xpath
        )
    )


class ThreadHTMLFetcher(object):

    def __init__(self, session, mailbox_number):
        self._session = session
        self._mailbox_number = mailbox_number

    def _query_params(self, start_at):
        return {
            'low': start_at,
            'folder': self._mailbox_number,
            'infiniscroll': 1
        }

    def fetch(self, start_at):
        response = self._session.okc_get('messages',
                                         params=self._query_params(start_at))
        return response.content.strip()

    def __repr__(self):
        return '{0}(mailbox_number={1})'.format(type(self).__name__,
                                                self._mailbox_number)


class MessageFetcher(object):

    def __init__(self, session, message_thread, read_messages=False):
        self._session = session
        self._read_messages = read_messages
        self._message_thread = message_thread

    @property
    def params(self):
        return {
            'readmsg': str(self._read_messages).lower(),
            'threadid': self._message_thread.id,
            'folder': 1
        }

    @util.cached_property
    def messages_tree(self):
        messages_response = self._session.okc_get('messages',
                                                  params=self.params)
        return html.fromstring(messages_response.content.decode('utf8'))

    def refresh(self):
        util.cached_property.bust_caches(self)
        return self.messages_tree

    def fetch(self):
        for message_element in self.message_elements:
            if message_element.attrib['id'] == 'compose':
                continue
            yield Message(message_element, self._message_thread)

    _message_elements_xpb = xpb.li.with_classes('to_me', 'from_me').or_

    @util.cached_property
    def message_elements(self):
        return self._message_elements_xpb.apply_(self.messages_tree)


class Message(object):

    def __init__(self, message_element, message_thread):
        self._message_element = message_element
        self._message_thread = message_thread

    @property
    def id(self):
        return int(self._message_element.attrib['id'].split('_')[-1])

    @util.cached_property
    def sender(self):
        return (self._message_thread.user_profile
                if 'from_me' in self._message_element.attrib['class']
                else self._message_thread.correspondent_profile)

    @util.cached_property
    def recipient(self):
        return (self._message_thread.correspondent_profile
                if 'from_me' in self._message_element.attrib['class']
                else self._message_thread.user_profile)

    @util.cached_property
    def content(self):
        message = xpb.div.with_class('message_body').apply_(self._message_element)
        content = None
        if message:
            message = message[0]
            content = message.text_content().replace(' \n \n', '\n').strip()
        return content

    def __repr__(self):
        return '<{0}: {1} sent {2} "{3}{4}">'.format(
            type(self).__name__,
            self.sender.username,
            self.recipient.username,
            self.content[:10],
            '...' if len(self.content) > 10 else ''
        )


class MessageThread(object):
    """Represent a message thread between two users."""

    def __init__(self, session, thread_element):
        self._session = session
        self._thread_element = thread_element
        self.reply = self.correspondent_profile.message(thread_id=self.id)
        self._message_fetcher = MessageFetcher(self._session, self)
        self.messages = util.Fetchable(self._message_fetcher)

    @util.cached_property
    def id(self):
        return self._thread_element.attrib['data-threadid']

    @util.cached_property
    def correspondent_id(self):
        return self._thread_element['data-personid']

    _correspondent_xpb = xpb.div.with_class('inner').a.with_class('photo').\
                             img.select_attribute_('alt')

    @util.cached_property
    def correspondent(self):
        return self._correspondent_xpb.apply_(self._thread_element)[0].split()[-1]

    @util.cached_property
    def read(self):
        return not 'unreadMessage' in self._thread_element.attrib['class']

    @util.cached_property
    def date(self):
        return self.datetime.date()

    @util.cached_property
    def datetime(self):
        timestamp_span = xpb.span.with_class('timestamp').apply_(self._thread_element)
        date_updated_text = timestamp_span[0][0].text_content()
        return helpers.parse_date_updated(date_updated_text)

    @property
    def with_deleted_user(self):
        try:
            self.correspondent_profile.id
        except exceptions.HTTPError:
            return True
        else:
            return False

    @property
    def initiator(self):
        try:
            return self.messages[0].sender
        except IndexError:
            pass

    @property
    def respondent(self):
        try:
            return self.messages[0].recipient
        except IndexError:
            pass

    @property
    def redact_messages(self):
        return self.restore(self.id, self.correspondent, self.read, self.date,
                            session=self._session,
                            messages=[self.redact_message(message)
                                      for message in self.messages])

    @staticmethod
    def redact_message(message):
        return Message(message.id, message.sender, message.recipient,
                       'x'*len(message.content))

    @util.cached_property
    def correspondent_profile(self):
        return self._session.get_profile(self.correspondent)

    @util.cached_property
    def user_profile(self):
        return self._session.get_current_user_profile()

    @property
    def refreshed_messages(self):
        if 'messages' in self.__dict__['messages']: del self.__dict__['messages']
        return self.messages

    @property
    def message_count(self):
        return len(self.messages)

    @property
    def has_messages(self):
        return bool(self.messages)

    @property
    def got_response(self):
        return any(message.sender != self.initiator for message in self.messages)

    def delete(self):
        return self._session.post('https://www.okcupid.com/mailbox',
                                  params=self.delete_params)

    @property
    def delete_params(self):
        return {
            'deletethread': 'DELETE',
            'mailaction': 3,
            'buddyname': self.correspondent,
            'r1': self.correspondent,
            'threadid': self.id,
            'receiverid': self.correspondent_id,
            'folderid': 1,
            'body_to_forward': self.messages[-1].content
        }

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return '<{0}({1}, {2})>'.format(
            type(self).__name__,
            self.user_profile.username,
            self.correspondent_profile.username
        )



# -*- coding: utf-8 -*-
import inspect
import logging
import re

from six import string_types

from . import helpers
from . import magicnumbers
from . import util
from .xpath import xpb


log = logging.getLogger(__name__)


def IndexedREMap(*re_strings, **kwargs):
    default = kwargs.get('default', 0)
    offset = kwargs.get('offset', 1)
    string_index_pairs = []
    for index, string_or_tuple in enumerate(re_strings, offset):
        if isinstance(string_or_tuple, string_types):
            string_or_tuple = (string_or_tuple,)
        for re_string in string_or_tuple:
            string_index_pairs.append((re_string, index))
    return util.REMap.from_string_pairs(string_index_pairs, default=default)


class Detail(object):

    NO_DEFAULT = object()

    @classmethod
    def comma_separated_presenter(cls, text):
        return text.strip().split(', ')

    @classmethod
    def mapping_multi_updater(cls, mapping):
        def updater(id_name, value):
            return {id_name: [mapping[item.lower()] for item in value]}
        return updater

    @classmethod
    def auto_indexed_updater(cls, *options):
        return cls.mapping_updater({option: index
                                    for index, option in enumerate(options, 1)},
                                   default=0)

    @classmethod
    def mapping_updater(cls, mapping, default=NO_DEFAULT):
        def updater(id_name, value):
            if isinstance(value, string_types):
                value = value.lower() # :/
            try:
                update = mapping[value]
            except KeyError:
                if default is not cls.NO_DEFAULT:
                    update = default
                else:
                    raise
            return {id_name: update}
        return updater

    @staticmethod
    def default_updater(id_name, value):
        return {id_name: value}

    def __init__(self, id_name=None, presenter=None, updater=None):
        self.id_name = id_name
        self.presenter = presenter or (lambda x: None if u'\u2014' in x
                                       else helpers.replace_chars(x.strip()))
        self.updater = updater or self.default_updater

    def update(self, value):
        if isinstance(value, string_types):
            value = value.lower()
        return self.updater(self.id_name, value)

    def __get__(self, details, klass):
        if details is None:
            return self
        return self.presenter(details.id_to_display_name_value[self.id_name])

    def __set__(self, details, value):
        details.update(self.update(value))


class DeclarativeDetail(object):

    updater = None
    presenter = None


class Details(object):

    @classmethod
    def name_detail_pairs(cls):
        is_detail = lambda x: isinstance(x, Detail)
        return inspect.getmembers(cls, is_detail)

    def __init__(self, profile):
        self.profile = profile

    _profile_details_xpb = xpb.div(id='profile_details').dl

    def refresh(self):
        util.cached_property.bust_caches(self)

    @util.cached_property
    def id_to_display_name_value(self):
        output = {}
        for element in self._profile_details_xpb.apply_(
            self.profile.profile_tree
        ):
            value_element = xpb.dd.one_(element)
            id_name = value_element.attrib['id'].replace('ajax_', '')
            value = value_element.text_content()
            output[id_name] = value
        return output

    @property
    def as_dict(self):
        return {name: getattr(self, name)
                for name, _ in self.name_detail_pairs()}

    def convert_and_update(self, data):
        klass = type(self)
        server_bound = {}
        for key, value in data.items():
            detail = getattr(klass, key)
            server_bound.update(detail.update(value))
        return self.update(server_bound)

    def update(self, data):
        log.debug(data)
        response = self.profile.authcode_post('profileedit2', data=data)
        self.profile.refresh()
        self.refresh()
        return response

    bodytype = Detail(updater=Detail.mapping_updater({
        None: 0, u'—': 0, 'rather not say': 1, 'thin': 2, 'overweight': 3,
        'skinny': 4, 'average': 5, 'fit': 6, 'athletic': 7, 'jacked': 8,
        'a little extra': 9, 'curvy': 10, 'full figured': 11, 'used up': 12
    }))

    orientation = Detail(updater=Detail.mapping_updater({
        'straight': 1, 'gay': 2, 'bisexual': 3
    }))

    ethnicities = Detail(
        presenter=Detail.comma_separated_presenter,
        updater=Detail.mapping_multi_updater(IndexedREMap(
            'asian', 'middle eastern', 'black', 'native american', 'indian',
            'pacific islander', ('hispanic', 'latin'), 'white', 'other'
        ))
    )

    smoking = Detail(updater=Detail.auto_indexed_updater(
        'yes', 'sometimes', 'when drinking', 'trying to quit', 'no'
    ))

    drugs = Detail(updater=Detail.mapping_updater(IndexedREMap(
        'never', 'sometimes', 'often', default=3, offset=0
    )))

    drinking = Detail(updater=Detail.auto_indexed_updater(
        'very often', 'often', 'socially', 'rarely', 'desperately', 'not at all'
    ))

    job = Detail(updater=Detail.mapping_updater(IndexedREMap(
        'student', ('art', 'music', 'writing'), ('banking', 'finance'),
        'administration', 'technology', 'construction', 'education',
        ('entertainment', 'media'), 'management', 'hospitality', 'law',
        'medicine', 'military', ('politics', 'government'),
        ('sales', 'marketing'), ('science', 'engineering'),
        'transportation', 'unemployed', 'other', 'rather not say',
        'retired'
    )))

    status = Detail(updater=Detail.mapping_updater(IndexedREMap(
        'single', 'seeing someone', 'married', 'in an open relationship'
    )))

    class monogamous(DeclarativeDetail):

        monogamous = IndexedREMap('(:?[^\-]monogamous)|(:?^monogamous)',
                                  'non-monogamous')

        strictness = IndexedREMap('mostly', 'strictly')

        @classmethod
        def updater(cls, id_name, value):
            return {
                'monogamous': cls.monogamous[value],
                'monogamyflex': cls.strictness[value]
            }

    class children(DeclarativeDetail):

        # Doesn't want kids is index 6 for some reason.
        has_kids = IndexedREMap('has a kid', 'has kids', (), (), (),
                                "doesn't have kids")
        wants_kids = IndexedREMap('might want', 'wants', "doesn't want")

        @classmethod
        def updater(cls, id_name, value):
            return {'children': cls.has_kids[value],
                    'children2': cls.wants_kids[value]}

    class education(DeclarativeDetail):

        education_status = util.REMap.from_string_pairs(
            (('[Gg]raduated', 1),
             ('[Ww]orking', 2),
             ('[Dd]ropped out', 3),
             ('[Ss]ome', 3)),
            default=0
        )


        education_level = util.REMap(((re.compile('high ?school'), 1),
                                      (re.compile('two[- ]year college'), 2),
                                      (re.compile('university'), 3),
                                      (re.compile('college'), 3),
                                      (re.compile('masters program'), 4),
                                      (re.compile('law school'), 5),
                                      (re.compile('med school'), 6),
                                      (re.compile('ph.d program'), 7),
                                      (re.compile('space camp'), 8)), default=0)

        @classmethod
        def updater(cls, id_name, value):
            value = value.lower()
            status = cls.education_status[value]
            level = cls.education_level[value]
            return {'educationstatus': status, 'educationlevel': level}

    class height(DeclarativeDetail):

        imperial_re = re.compile(u"([0-9])['′\u2032] ?([0-9][0-1]?)[\"″\u2033]",
                                 flags=re.UNICODE)
        metric_re = re.compile(u"([0-2]\.[0-9]{2})m")

        @classmethod
        def updater(cls, id_name, value):
            match = cls.imperial_re.search(value)
            if match:
                return {'feet': match.group(1),
                        'inches': match.group(2)}

            match = cls.metric_re.search(value)
            if match:
                meters = float(match.group(1))
                centimeters = meters * 100
                return {'centimeters': int(centimeters)}
            else:
                raise ValueError("The provided height did not match any of "
                                 "the accepted formats.")

    class diet(DeclarativeDetail):

        diet_strictness = IndexedREMap('[Mm]ostly', '[Ss]trictly')

        diet_type = IndexedREMap('anything', 'vegetarian', 'vegan',
                                  'kosher', 'halal', 'other')

        @classmethod
        def updater(cls, id_name, value):
            return {'diet': cls.diet_type[value],
                    'dietserious': cls.diet_strictness[value]}

    class religion(DeclarativeDetail):

        religion = IndexedREMap('agnosticism', 'atheism', 'christianity',
                                'judaism', 'catholicism', 'islam', 'hinduism',
                                'buddhism', 'other', default=1, offset=2)
        seriousness = IndexedREMap('very serious', 'somewhat serious',
                                   'not too serious', 'laughing')

        @classmethod
        def updater(cls, id_name, value):
            return {'religion': cls.religion[value],
                    'religionserious': cls.seriousness[value]}

    class sign(DeclarativeDetail):

        sign = IndexedREMap(
            'aquarius', 'pisces', 'aries', 'taurus', 'gemini', 'cancer', 'leo',
            'virgo', 'libra', 'scorpio', 'sagittarius', 'capricorn',
        )

        importance = IndexedREMap(
            "doesn't matter", "matters alot", "fun to think about"
        )

        @classmethod
        def updater(cls, id_name, value):
            return {'sign': cls.sign[value],
                    'sign_status': cls.importance[value]}

    class income(DeclarativeDetail):

        levels = list(enumerate(
            (20000, 30000, 40000, 50000, 60000, 70000,
             80000, 100000, 150000, 250000, 500000, 1000000)
        ))

        comma_sep_number = '([0-9]{1,3}(?:,[0-9]{3})*)'

        range_matcher = re.compile(u'\$?{0}-\$?{0}'.format(comma_sep_number),
                                   flags=re.UNICODE)
        lt_matcher = re.compile("less than \$?{0}".format(comma_sep_number),
                                flags=re.UNICODE)
        gt_matcher = re.compile("more than \$?{0}".format(comma_sep_number),
                                flags=re.UNICODE)

        @classmethod
        def updater(cls, id_name, value):
            if value is None:
                return {'income': 0}
            if isinstance(value, string_types):
                for matcher, sign in ((cls.range_matcher, 1),
                                      (cls.lt_matcher, -1),
                                      (cls.gt_matcher, 1)):
                    match = matcher.match(value)
                    if match:
                        matched_income = int(match.group(1).replace(',', ''))
                        value = matched_income + 100 * sign
                        break
            for index, level in cls.levels:
                if value < level:
                    break
            else:
                index += 1
            update = index + 1
            return {'income': update}

    class pets(DeclarativeDetail):

        dogs = util.REMap.from_string_pairs(
            (('dislikes dogs', 3), ('has dogs', 1), ('likes dogs', 2)),
            default=0
        )

        cats = util.REMap.from_string_pairs(
            (('dislikes cats', 3), ('has cats', 1), ('likes cats', 2)),
            default=0
        )

        @classmethod
        def updater(cls, id_name, value):
            return {'cats': cls.cats[value],
                    'dogs': cls.dogs[value]}


    class languages(DeclarativeDetail):

        language_matcher = re.compile('(.*?) \((.*?)\)')

        language_to_number = magicnumbers.language_map_2

        level = {
            'fluently': 1, 'okay': 2, 'poorly': 3, None: 0
        }

        @classmethod
        def presenter(cls, value):
            language_strings = value.split(',')
            languages = []
            for language_string in language_strings:
                match = cls.language_matcher.match(language_string.strip())
                languages.append((match.group(1).lower(),
                                  match.group(2).lower()))
            return languages

        @classmethod
        def updater(cls, id_name, languages):
            data = {}
            for number, (language, level) in enumerate(languages, 1):
                language_number = cls.language_to_number[language.lower()]
                level_number = cls.level[level.lower()]
                data['cont_lang_{0}'.format(number)] = language_number
                data['language{0}status'.format(number)] = level_number
            number += 1
            for i in range(number, 6):
                data['cont_lang_{0}'.format(i)] = ''
                data['language{0}status'.format(i)] = ''
            return data


for id_name, detail in Details.name_detail_pairs():
    if detail.id_name is None:
        detail.id_name = id_name

    is_declarative_detail = lambda x: (isinstance(x, type) and
                                  issubclass(x, DeclarativeDetail))
    for id_name, declarative_detail in inspect.getmembers(
        Details, is_declarative_detail
    ):
        detail = Detail(presenter=declarative_detail.presenter,
                        updater=declarative_detail.updater,
                        id_name=id_name)
        setattr(Details, id_name, detail)



"""Print a random profile from the quickmatch page."""

import okcupyd

u = okcupyd.User()
p = u.quickmatch() # p is an instance of the Profile class

print('Profile of {0}'.format(p.username))

print('\n'.join('{0}{1}'.format(title, value)
                for title, value in (('Gender', p.gender),
                                     ('Age', p.age),
                                     ('Orientation', p.orientation),
                                     ('Location', p.location),
                                     ('Match', p.match_percentage),
                                     ('Enemy', p.enemy_percentage))))



import os


USERNAME = os.environ.get('OKC_USERNAME')
PASSWORD = os.environ.get('OKC_PASSWORD')

AF_USERNAME = os.environ.get('AF_USERNAME', USERNAME)
AF_PASSWORD = os.environ.get('AF_PASSWORD', PASSWORD)



import mock
import pytest
from sqlalchemy.orm.exc import NoResultFound

from okcupyd import db
from okcupyd.db import model


def test_integrity_error_on_okc_id_handled_by_safe_upsert():
    session = db.Session()
    other_session = db.Session()
    m1 = model.MessageThread(okc_id=1, initiator_id=1, respondent_id=1)
    m2 = model.MessageThread(okc_id=1, initiator_id=1, respondent_id=2)

    other_session.add(m1)
    find_all_no_txn = model.MessageThread.find_all_no_txn
    def commit_other_session(*args, **kwargs):
        other_session.commit()
        model.MessageThread.find_all_no_txn = find_all_no_txn
        return []
        with mock.patch.object(model.MessageThread, 'find_all_no_txn',
                               commit_other_session):
            model.MessageThread.safe_upsert([m2], id_key='okc_id')

    other_session.commit()
    session.commit()


def test_model_missing():
    with pytest.raises(NoResultFound):
        model.MessageThread.find(24)


def test_nest_txn():
    with db.txn() as session:
        session.add(model.MessageThread(okc_id=2, initiator_id=1, respondent_id=1))
        with db.txn() as nested:
            nested.add(model.MessageThread(okc_id=1, initiator_id=1, respondent_id=1))
            assert nested is not session
            session.add(model.MessageThread(okc_id=3, initiator_id=1, respondent_id=1))


def test_upsert():
    new_user_1 = model.User(handle='fun', okc_id='1',
                            age=40, location='Washington, D.C.')
    new_user_2 = model.User(handle='other', okc_id='1',
                            age=40, location='Washington, D.C.')
    v1 = model.User.upsert_okc(new_user_1)
    v2 = model.User.upsert_okc(new_user_2)

    loaded_from_v1 = model.User.find(v1.id)
    assert loaded_from_v1.handle == v2.handle
    assert v2.handle == new_user_2.handle
    assert v2.handle == 'other'
    assert len(model.User.query()) == 1


import os

import pytest

from . import util
from okcupyd import User
from okcupyd import helpers
from okcupyd.profile import Profile


@util.use_cassette(cassette_name='user_no_picture')
def test_handle_no_pictures():
    username = User().profile.username
    assert username is not None


@util.use_cassette(cassette_name='access_profile_from_message_thread')
def test_message_thread_to_profile():
    profile = User().inbox[0].correspondent_profile
    assert profile.age
    assert profile.age > 18
    assert isinstance(profile.rating, int)


@util.use_cassette(cassette_name='user_count')
def test_user_search_count():
    assert len(User().search(count=1)) == 1


@util.use_cassette(cassette_name='test_user_essays')
def test_user_essays():
    user = User()
    first_essay = 'an essay'
    user.profile.essays.self_summary = first_essay
    assert user.profile.essays.self_summary == first_essay

    second_essay = 'next_essay'
    user.profile.essays.self_summary = second_essay

    assert user.profile.essays.self_summary == second_essay


@util.use_cassette(cassette_name='test_user_essays_refresh')
def test_user_essay_refresh():
    # Test Refresh Function
    user = User()
    user2 = User(user._session)
    user.message_me_if = 'other stuff'

    user2.profile.essays.refresh()
    assert user.profile.essays.message_me_if == user2.profile.essays.message_me_if


@util.use_cassette(cassette_name='visitors_test')
def test_visitors():
    user = User()
    assert isinstance(user.visitors[0], Profile)
    # If the user has more than one page of visitors, this ensures that
    assert isinstance(user.visitors[-1], Profile)
    assert len(user.visitors) > 26


@pytest.mark.skipif(bool(os.environ.get('CI')), reason="Unicode issues...")
@util.use_cassette(cassette_name='profile_titles')
def test_profile_titles():
    user = User()
    for essay_name in user.profile.essays.essay_names:
        setattr(user.profile.essays, essay_name, 'updated')

    expected_names_dict = {
        'favorites': 'Favorite books, movies, shows, music, and food',
        'friday_night': 'On a typical Friday night I am',
        'good_at': u'I\u2019m really good at',
        'message_me_if': 'You should message me if',
        'my_life': u'What I\u2019m doing with my life',
        'people_first_notice': 'The first things people usually notice about me',
        'private_admission': u'The most private thing I\u2019m willing to admit',
        'self_summary': 'My self-summary',
        'six_things': 'The six things I could never do without',
        'think_about': 'I spend a lot of time thinking about'
    }

    expected_names_dict = {key: helpers.replace_chars(value)
                           for key, value in expected_names_dict.items()}

    assert user.profile.essays.short_name_to_title == expected_names_dict


@util.use_cassette
def test_user_profile_attributes():
    user_profile = User().profile
    assert user_profile.id > 0
    assert user_profile.rating == 0



"""The public interface to `search` is through `SearchFetchable`.

profiles = SearchFetchable(location='new york, ny', religion='buddhist',
                           height_min=66, height_max=68, gentation='everybody',
                           smokes=['no', 'trying to quit'])
      for profile in profiles[:2]:
        for photo_info in profile.photo_infos:
          print photo_info.jpg_uri
"""

import logging

import simplejson

from . import helpers
from . import magicnumbers
from . import util
from . import filter
from .profile import Profile
from .session import Session
from .xpath import xpb


log = logging.getLogger(__name__)


search_filters = filter.Filters()
search_filters.register_filter_builder(
    filter.gentation_filter,
    descriptions="The gentation of returned search results.",
    acceptable_values=magicnumbers.gentation_to_number.keys(),
    types=str
)
search_filters.register_filter_builder(
    filter.location_filter,
    types=int,
    descriptions=("The maximum distance from the specified location of "
                  "returned search results.")
)
search_filters.register_filter_builder(
    filter.age_filter,
    decider=filter.Filters.any_not_none_decider,
    types=int,
    descriptions=['The minimum age of returned search results.',
                  'The maximum age of returned search results.']
)


@search_filters.register_filter_builder(
    decider=search_filters.any_decider,
    types=int,
    descriptions=["The minimum attractiveness of returned search results.",
                  "The maximum attractiveness of returned search results."]
)
def attractiveness_filter(attractiveness_min, attractiveness_max):
    if attractiveness_min == None:
        attractiveness_min = 0
    if attractiveness_max == None:
        attractiveness_max = 10000
    return '25,{0},{1}'.format(attractiveness_min, attractiveness_max)


@search_filters.register_filter_builder(
    decider=search_filters.any_not_none_decider
)
def height_filter(height_min, height_max):
    return magicnumbers.get_height_query(height_min, height_max)


@search_filters.register_filter_builder
def last_online_filter(last_online):
    return '5,{0}'.format(helpers.format_last_online(last_online))


@search_filters.register_filter_builder
def status_filter(status):
    status_int = 2  # single, default
    if status.lower() in ('not single', 'married'):
        status_int = 12
    elif status.lower() == 'any':
        status_int = 0
    return '35,{0}'.format(status_int)


def build_option_filter(key):
    @search_filters.register_filter_builder(
        keys=(key,),
        acceptable_values=magicnumbers.binary_lists[key],
        types=str
    )
    @util.makelist_decorator
    def option_filter(value):
        return magicnumbers.get_options_query(key, value)


for key in ['smokes', 'drinks', 'drugs', 'education', 'job',
            'income', 'religion', 'monogamy', 'diet', 'sign',
            'ethnicity']:
    build_option_filter(key)


         # ('offspring', util.makelist_decorator(magicnumbers.get_kids_query)),
         # ('join_date', magicnumbers.get_join_date_query),
         # ('languages', magicnumbers.get_language_query)]

# search_filters.register_filter_builder(
#     util.makelist_decorator(magicnumbers.get_pet_queries)),
# keys=('pets',),
#     types=str,
# )


_username_xpb = xpb.div.with_classes('match_card').\
                div.with_class('username').a.text_
def SearchFetchable(session=None, **kwargs):
    """Search okcupid.com with the given parameters.

    :returns: A :class:`okcupyd.util.fetchable.Fetchable` of
              :class:`okcupyd.profile.Profile` instances.
    """

    session = session or Session.login()
    return util.Fetchable.fetch_marshall(
        SearchHTMLFetcher(session, **kwargs),
        util.SimpleProcessor(
            session,
            lambda username: Profile(session, username.strip()),
            _username_xpb
        )
    )


SearchFetchable.__doc__ += '\n' + search_filters.build_documentation()


class SearchHTMLFetcher(object):

    _username_xpb = xpb.div.with_class('username')

    def __init__(self, session=None, **options):
        self._session = session or Session.login()
        self._options = options

    @property
    def location(self):
        return self._options.get('location', None)

    @property
    def gender(self):
        return self._options.get('gender', 'm')

    @property
    def keywords(self):
        return self._options.get('keywords')

    @property
    def order_by(self):
        return self._options.get('order_by', 'match').upper()

    def _query_params(self, count=None, low=None):
        search_parameters = {
            'timekey': 1,
            'matchOrderBy': self.order_by,
            'custom_search': '0',
            'fromWhoOnline': '0',
            'mygender': self.gender,
            'update_prefs': '1',
            'sort_type': '0',
            'sa': '1',
            'count': str(count) if count else self._options.get('count', 9),
            'locid': (str(helpers.get_locid(self._session, self.location))
                      if self.location else 0),
            'ajax_load': 1,
            'discard_prefs': 1,
            'match_card_class': 'just_appended'
        }
        if low:
            search_parameters['low'] = low
        if self.keywords: search_parameters['keywords'] = self.keywords
        search_parameters.update(search_filters.build(**self._options))
        return search_parameters

    def fetch(self, start_at=None, count=None):
        search_parameters = self._query_params(low=start_at, count=count)
        log.info(simplejson.dumps({'search_parameters': search_parameters}))
        response = self._session.okc_get('match',
                                         params=search_parameters)
        try:
            search_html = response.json()['html']
        except:
            log.warning(simplejson.dumps({'failure': response.content}))
            raise
        return search_html

    def __unicode__(self):
        return u'{0}({1})'.format(type(self).__name__, repr(self._options))

    __repr__ = __unicode__


def search(session=None, count=1, **kwargs):
    return SearchFetchable(session, count=count, **kwargs)[:count]



from . import util
from okcupyd import photo


@util.use_cassette(cassette_name='photo_upload',
                   match_on=util.match_on_no_body)
def test_photo_upload():
    uploader = photo.PhotoUploader()
    upload_response_dict = uploader.upload_and_confirm('fixtures/image.jpg')
    assert int(upload_response_dict['id']) > 0



import logging

import simplejson

from okcupyd import helpers
from okcupyd.db import adapters
from okcupyd.db import model, txn


log = logging.getLogger(__name__)


class Sync(object):

    def __init__(self, user):
        self._user = user

    def all(self):
        with txn() as session:
            okcupyd_user = session.query(model.OKCupydUser).join(model.User).filter(
                model.User.okc_id == self._user.profile.id
            ).with_for_update().one()

            log.info(simplejson.dumps({
                'outbox_last_updated': helpers.datetime_to_string(
                    okcupyd_user.outbox_last_updated
                ),
                'inbox_last_updated': helpers.datetime_to_string(
                    okcupyd_user.inbox_last_updated
                )
            }))

            okcupyd_user.outbox_last_updated = self._sync_mailbox_until(
                self._user.outbox(),
                okcupyd_user.outbox_last_updated
            )
            okcupyd_user.inbox_last_updated = self._sync_mailbox_until(
                self._user.inbox(),
                okcupyd_user.inbox_last_updated
            )

    def _sync_mailbox_until(self, mailbox, sync_until):
        for thread in mailbox:
            if sync_until and sync_until > thread.datetime:
                break
            if not thread.messages:
                continue
            if not thread.with_deleted_user:
                adapters.ThreadAdapter(thread).get_thread()
        try:
            return mailbox[0].datetime
        except IndexError:
            pass



import importlib
import os

import mock

from . import util


path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'examples')


def build_example_test_function(module_name):
    @mock.patch('okcupyd.save_file')
    def function(arg):
        with util.use_cassette(cassette_name='{0}_example'.format(module_name)):
            importlib.import_module('examples.{0}'.format(module_name))
    return function


for root, _, filenames in os.walk(path):
    for filename in filenames:
        try:
            module_name, extention = filename.split('.')
        except:
            continue
        if extention != 'py' or module_name == '__init__':
            continue
        function = build_example_test_function(module_name)
        locals()['test_{0}_example'.format(module_name)] = function



import itertools
import operator

import mock
import pytest

from okcupyd import util


@util.curry
def crazy(a, self):
    return a + self.starter


class HasCurryMethods(object):

    def __init__(self, starter):
        self.starter = starter
        self.on_instance = crazy

    @util.curry
    def curry(self, x, y=0):
        return self.starter * x * y

    on_class = crazy(1)


def test_curry_method_behavior():
    assert HasCurryMethods(2).curry(y=2)(1) == 4
    hpm = HasCurryMethods(2)
    assert hpm.on_instance(2, hpm) == 4
    assert hpm.on_class() == 3


def test_curry_with_kwargs_taking_function():
    @util.curry
    def kwarg_taking_function(arg, **kwargs):
        kwargs['k'] = arg
        return kwargs

    assert kwarg_taking_function(a=14)(2) == {'k': 2, 'a': 14}


def test_get_cached_properties():
    class PropClass(object):
        def __init__(self):
            self.count = 0

        @util.cached_property
        def count_prop(self):
            self.count += 1
            return self.count

        @util.cached_property
        def other_prop(self):
            return 4

    instance = PropClass()
    assert ['count_prop', 'other_prop'] == \
        list(map(operator.itemgetter(0),
                 util.cached_property.get_cached_properties(instance)))

    assert instance.count_prop == 1

    util.cached_property.bust_caches(instance)

    assert instance.count_prop == 2


def test_overwrite_kwarg():

    @util.curry
    def test(k, a=0, b=0):
        return k + a + b

    assert test(a=4)(a=1)(1) == 2


def test_fetchable_teeing():
    fetcher = mock.Mock(fetch=lambda: itertools.count(5))
    fetchable = util.Fetchable(fetcher)
    first = iter(fetchable)
    second = iter(fetchable)

    assert next(first) == 5
    assert next(second) == 5

    third = iter(fetchable)

    assert next(third) == 5
    assert next(third) == 6
    assert next(third) == 7

    assert next(first) == 6
    assert next(second) == 6

    for i in range(4):
        assert next(iter(fetchable)) == 5

def test_fetchable_get_item():
    call_counter = mock.Mock()
    def fetch():
        for i in range(3):
            for i in range(5):
                yield i
            call_counter()
    fetcher = mock.Mock(fetch=fetch)
    fetchable = util.Fetchable(fetcher)

    assert fetchable[:2] == [0, 1]
    assert fetchable[:3] == [0, 1, 2]
    assert fetchable[3] == 3
    assert call_counter.call_count == 0

    assert fetchable[9] == 4
    assert call_counter.call_count == 1

    assert fetchable[2:3] == [2]

    assert fetchable[:] == fetchable[:]
    assert len(fetchable[:]) == 15

    # Go too far on purpose
    assert fetchable[:100000] == fetchable[:]

    # Go too far on purpose
    assert fetchable[100000:] == []
    assert fetchable[100000:10000000] == []

    with pytest.raises(IndexError):
        fetchable[12312312]


def test_negative_indexing():
    def fetch():
        yield 0
        yield 0
        yield 1
    fetcher = mock.Mock(fetch=fetch)
    fetchable = util.Fetchable(fetcher)
    assert fetchable[-1] == 1
    assert fetchable[-2:] == [0, 1]
    assert fetchable[14:] == []


def test_bool_on_fetchable():
    fetcher = mock.Mock(fetch=lambda: (i for i in range(0)))
    fetchable = util.Fetchable(fetcher)
    assert not fetchable

    fetcher = mock.Mock(fetch=lambda: (i for i in range(1)))
    fetchable = util.Fetchable(fetcher)

    assert fetchable


def test_curry_on_classmethod():
    class TestClass(object):

        @classmethod
        @util.curry
        def test(cls, arg, optional=3):
            return optional + arg

    assert TestClass.test(optional=1)(2) == 3
    assert TestClass().test(optional=3)(2) == 5



import mock
import pytest

from . import util
from okcupyd.attractiveness_finder import _AttractivenessFinder, AttractivenessFinder


@pytest.fixture
def mock_session():
    return mock.Mock()


@pytest.fixture
def attractiveness_finder(mock_session):
    return _AttractivenessFinder(mock_session)


@mock.patch('okcupyd.attractiveness_finder.search')
def test_attractiveness_finder_stops(mock_search, attractiveness_finder):
    assert attractiveness_finder.find_attractiveness('test_user', accuracy=10000)
    assert mock_search.call_count == 0


@mock.patch('okcupyd.attractiveness_finder.search')
def test_attractiveness_finder(mock_search, attractiveness_finder):
    user_one = 'user_one'
    user_two = 'user_two'
    user_to_attractiveness = {
        user_one: 4875,
        user_two: 9212
    }
    def mock_search_function(session, attractiveness_min=0,
                    attractiveness_max=10000, keywords='', **kwargs):
        if attractiveness_min <= user_to_attractiveness[keywords] <= attractiveness_max:
            return mock.Mock()
    mock_search.side_effect = mock_search_function

    user_one_attractiveness = attractiveness_finder(user_one, accuracy=1)
    assert user_one_attractiveness == user_to_attractiveness[user_one]

    user_two_attractiveness = attractiveness_finder.find_attractiveness(user_two,
                                                                        accuracy=1)
    assert user_two_attractiveness == user_to_attractiveness[user_two]

    # Test Rounding
    attractiveness_finder = AttractivenessFinder(mock.Mock())

    assert attractiveness_finder(user_one) == 5000
    assert attractiveness_finder(user_two) == 9000


@pytest.yield_fixture
def cached_attractiveness_finder():
    with util.use_cassette(cassette_name='attractiveness_finder'):
        yield AttractivenessFinder()


@util.use_cassette(cassette_name='attractiveness_finder_live')
def test_attractiveness_finder_live(cached_attractiveness_finder):
    assert cached_attractiveness_finder('_iWarhol') == 1000
    assert cached_attractiveness_finder('dasmitches') == 5000



from . import util
from . import helpers
from .xpath import xpb


class Essays(object):
    """Interface to reading and writing a users essays."""

    @staticmethod
    def build_essay_property(essay_index, essay_name):
        essay_xpb = xpb.div(id='essay_{0}'.format(essay_index))
        essay_text_xpb = essay_xpb.div.with_class('text').div.with_class('essay')
        @property
        def essay(self):
            try:
                essay_text = essay_text_xpb.get_text_(self._essays).strip()
            except IndexError:
                return None
            if essay_name not in self._short_name_to_title:
                self._short_name_to_title[essay_name] = helpers.replace_chars(
                    essay_xpb.a.with_class('essay_title').get_text_(
                        self._profile.profile_tree
                    )
                )
            return essay_text

        @essay.setter
        def set_essay_text(self, essay_text):
            self._submit_essay(essay_index, essay_text)

        return set_essay_text

    @classmethod
    def _init_essay_properties(cls):
        for essay_index, essay_name in enumerate(cls.essay_names):
            setattr(cls, essay_name,
                    cls.build_essay_property(essay_index, essay_name))

    _essays_xpb = xpb.div(id='main_column')
    #: A list of the attribute names that are used to store the text of
    #: of essays on instances of this class.
    essay_names = ['self_summary', 'my_life', 'good_at', 'people_first_notice',
                   'favorites', 'six_things', 'think_about', 'friday_night',
                   'private_admission', 'message_me_if']

    def __init__(self, profile):
        """:param profile: A :class:`.Profile`"""
        self._profile = profile
        self._short_name_to_title = {}

    @property
    def short_name_to_title(self):
        for i in self: pass # Make sure that all essays names have been retrieved
        return self._short_name_to_title

    @util.cached_property
    def _essays(self):
        return self._essays_xpb.one_(self._profile.profile_tree)

    def _submit_essay(self, essay_id, essay_body):
        self._profile.authcode_post('profileedit2', data={
            "essay_id": essay_id,
            "essay_body": essay_body,
            "okc_api": 1
        })
        self.refresh()

    def refresh(self):
        self._profile.refresh()
        util.cached_property.bust_caches(self)

    def __iter__(self):
        for essay_name in self.essay_names:
            yield getattr(self, essay_name)


Essays._init_essay_properties()



# -*- coding: utf-8 -*-
from collections import namedtuple
from datetime import datetime, timedelta
from json import loads
from lxml import html
from re import search
import logging

import simplejson

from .xpath import xpb
from . import util


CHAR_REPLACE = {
    u"′": "'",
    u'″': '"',
    u'“': '"',
    u'”': '"',
    u"’": "'",
    u"—": "-",
    u"–": "-",
    u"…": "...",
    u"🌲": " ",
    u'\u2019': "'",
}


log = logging.getLogger(__name__)


MessageInfo = namedtuple('MessageInfo', ('thread_id', 'message_id'))


class Messager(object):

    def __init__(self, session):
        self._session = session

    def _get_authcode(self, username):
        response = self._session.get(
            'https://www.okcupid.com/profile/{0}'.format(username)
        ).content.decode('utf8')
        return get_authcode(response)

    def message_request_parameters(self, username, message, thread_id, authcode):
        return {
            'ajax': 1,
            'sendmsg': 1,
            'r1': username,
            'body': message,
            'threadid': thread_id,
            'authcode': authcode,
            'reply': 1 if thread_id else 0,
            'from_profile': 1
        }

    def send(self, username, message, authcode=None, thread_id=None):
        authcode = authcode or self._get_authcode(username)
        params = self.message_request_parameters(username, message,
                                                 thread_id or 0, authcode)
        response = self._session.get('https://www.okcupid.com/mailbox', params=params)
        response_dict = response.json()
        log.info(simplejson.dumps({'message_send_response': response_dict}))
        return MessageInfo(response_dict.get('threadid'), response_dict['msgid'])


@util.curry
def get_js_variable(html_response, variable_name):
    script_elements = xpb.script.apply_(html_response)
    html_response = '\n'.join(script_element.text_content()
                              for script_element in script_elements)
    return search('var {0} = "(.*?)";'.format(variable_name), html_response).group(1)


get_authcode = get_js_variable(variable_name='AUTHCODE')
get_username = get_js_variable(variable_name='SCREENNAME')
get_id = get_js_variable(variable_name='CURRENTUSERID')
get_current_user_id = get_id
get_my_username = get_username


weekday_to_ordinal = {
    'monday': 0,
    'tuesday': 1,
    'wednesday': 2,
    'thursday': 3,
    'friday': 4,
    'saturday': 5,
    'sunday': 6
}


def parse_date_updated(date_updated_text):
    recognized = False
    for function in (parse_slashed_date, parse_abbreviated_date,
                     parse_time, parse_day_of_the_week, parse_contextual_date):
        parsed_time = function(date_updated_text)
        if parsed_time is not None:
            recognized = True
            break
    else:
        parsed_time = datetime.today()
    log_function = log.info if recognized else log.error
    log_function(simplejson.dumps({"incoming_date": date_updated_text,
                                   "outgoing_date": parsed_time.strftime("%Y.%m.%d %H:%M"),
                                   "recognized": recognized}))
    return parsed_time


def parse_contextual_date(date_updated_text):
    if 'yesterday' in date_updated_text.lower():
        return datetime.now() - timedelta(days=1)
    if 'now' in date_updated_text.lower():
        return datetime.now()


def parse_slashed_date(date_updated_text):
    try:
        return datetime.strptime(date_updated_text, '%m/%d/%y')
    except:
        pass

def parse_abbreviated_date(date_updated_text):
    try:
        parsed_time = datetime.strptime(date_updated_text, '%b %d')
    except:
        pass
    else:
        return parsed_time.replace(year=datetime.now().year)

def parse_time(date_updated_text):
    try:
        time = datetime.strptime(date_updated_text, '%I:%M%p')
    except:
        pass
    else:
        today = datetime.today()
        parsed_time = time.replace(year=today.year, day=today.day,
                                   month=today.month)
        if parsed_time > datetime.now():
            parsed_time = parsed_time - timedelta(days=1)
        return parsed_time


def parse_day_of_the_week(date_updated_text):
    try:
        datetime.strptime(date_updated_text, '%A')
    except:
        pass
    else:
        return date_from_weekday(date_updated_text)


def date_from_weekday(weekday):
    today = datetime.now()
    incoming_weekday_ordinal = weekday_to_ordinal[weekday.lower()]
    today_ordinal = today.weekday()
    difference = (today_ordinal - incoming_weekday_ordinal
                  if today_ordinal > incoming_weekday_ordinal else
                  7 - incoming_weekday_ordinal + today_ordinal)
    return today - timedelta(days=difference)


def datetime_to_string(a_datetime):
    if a_datetime:
        return a_datetime.strftime('%H:%M:%S')


def get_locid(session, location):
    """
    Make a request to locquery resource to translate a string location
    search into an int locid.
    Returns
    ----------
    int
        An int that OKCupid maps to a particular geographical location.
    """
    locid = 0
    query_parameters = {
        'func': 'query',
        'query': location,
    }
    loc_query = session.get('http://www.okcupid.com/locquery', params=query_parameters)
    p = html.fromstring(loc_query.content.decode('utf8'))
    js = loads(p.text)
    if 'results' in js and len(js['results']):
        locid = js['results'][0]['locid']
    return locid


def format_last_online(last_online):
    """
    Return the upper limit in seconds that a profile may have been
    online. If last_online is an int, return that int. Otherwise if
    last_online is a str, convert the string into an int.
    Returns
    ----------
    int
    """
    if isinstance(last_online, str):
        if last_online.lower() in ('day', 'today'):
            last_online_int = 86400  # 3600 * 24
        elif last_online.lower() == 'week':
            last_online_int = 604800  # 3600 * 24 * 7
        elif last_online.lower() == 'month':
            last_online_int = 2678400  # 3600 * 24 * 31
        elif last_online.lower() == 'year':
            last_online_int = 31536000  # 3600 * 365
        elif last_online.lower() == 'decade':
            last_online_int = 315360000  # 3600 * 365 * 10
        else: # Defaults any other strings to last hour
            last_online_int = 3600
    else:
        last_online_int = last_online
    return last_online_int


def update_looking_for(profile_tree, looking_for):
    """
    Update looking_for attribute of a Profile.
    """
    div = profile_tree.xpath("//div[@id = 'what_i_want']")[0]
    looking_for['gentation'] = div.xpath(".//li[@id = 'ajax_gentation']/text()")[0].strip()
    looking_for['ages'] = replace_chars(div.xpath(".//li[@id = 'ajax_ages']/text()")[0].strip())
    looking_for['near'] = div.xpath(".//li[@id = 'ajax_near']/text()")[0].strip()
    looking_for['single'] = div.xpath(".//li[@id = 'ajax_single']/text()")[0].strip()
    try:
        looking_for['seeking'] = div.xpath(".//li[@id = 'ajax_lookingfor']/text()")[0].strip()
    except:
        pass


def update_details(profile_tree, details):
    """
    Update details attribute of a Profile.
    """
    div = profile_tree.xpath("//div[@id = 'profile_details']")[0]
    for dl in div.iter('dl'):
        title = dl.find('dt').text
        item = dl.find('dd')
        if title == 'Last Online' and item.find('span') is not None:
            details[title.lower()] = item.find('span').text.strip()
        elif title.lower() in details and len(item.text):
            details[title.lower()] = item.text.strip()
        else:
            continue
        details[title.lower()] = replace_chars(details[title.lower()])


gender_to_orientation_to_gentation = {
    'm': {
        'straight': 'girls who like guys',
        'gay': 'guys who like guys',
        'bisexual': 'both who like bi guys'
    },
    'f': {
        'straight': 'guys who like girls',
        'gay': 'girls who like girls',
        'bisexual': 'both who like bi girls'
    }
}


def get_default_gentation(gender, orientation):
    """Return the default gentation for the given gender and orientation."""
    gender = gender.lower()[0]
    orientation = orientation.lower()
    return gender_to_orientation_to_gentation[gender][orientation]


def replace_chars(astring):
    """
    Replace certain unicode characters to avoid errors when trying
    to read various strings.
    Returns
    ----------
    str
    """
    for k, v in CHAR_REPLACE.items():
        astring = astring.replace(k, v)
    return astring


def add_newlines(tree):
    """
    Add a newline character to the end of each <br> element.
    """
    for br in tree.xpath("*//br"):
        br.tail = "\n" + br.tail if br.tail else "\n"



from setuptools import setup, find_packages


setup(
    name="okcupyd",
    version="0.8.0",
    packages=find_packages(exclude=('tests*', 'examples')),
    install_requires=['lxml', 'requests >= 2.4.1', 'simplejson',
                      'sqlalchemy >= 0.9.0', 'ipython >= 2.2.0',
                      'wrapt', 'coloredlogs >= 0.8', 'invoke >= 0.9', 'six'],
    tests_require=['tox', 'pytest', 'mock', 'contextlib2', 'vcrpy'],
    package_data={'': ['*.md', '*.rst']},
    author="Ivan Malison",
    author_email="ivanmalison@gmail.com",
    description="A package for interacting with okcupid.com",
    license="MIT",
    keywords=["okcupid", "okcupyd", "pyokc" "online dating"],
    url="https://github.com/IvanMalison/okcupyd",
    long_description='',
    entry_points={"console_scripts": ["okcupyd=okcupyd:interactive"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
)



import logging

from invoke import task
import IPython

from okcupyd import db
from okcupyd import util
from okcupyd.db import mailbox, model
from okcupyd.user import User


log = logging.getLogger(__name__)


@task(default=True)
def session():
    with db.txn() as session:
        IPython.embed()


@task
def reset():
    util.enable_logger(__name__)
    log.info(db.Base.metadata.bind)
    db.Base.metadata.drop_all()
    db.Base.metadata.create_all()


@task
def sync():
    user = User()
    mailbox.Sync(user).all()
    log.info(model.Message.query(model.User.okc_id == user.profile.id))


@task
def make():
    user = User()
    user_model = model.User.from_profile(user.profile)
    user_model.upsert_model(id_key='okc_id')
    okcupyd_user = model.OKCupydUser(user_id=user_model.id)
    okcupyd_user.upsert_model(id_key='user_id')
    return okcupyd_user



import logging

from . import util
from .xpath import xpb


log = logging.getLogger(__name__)


class BaseQuestion(object):

    def __init__(self, question_element):
        self._question_element = question_element

    @property
    def answered(self):
        return 'not_answered' not in self._question_element.attrib['class']

    @util.cached_property
    def id(self):
        return self._question_element.attrib['data-qid']

    _text_xpb = xpb.div.with_class('qtext').p

    @util.cached_property
    def text(self):
        return self._text_xpb.get_text_(self._question_element).strip()

    def __repr__(self):
        return '<Question: {0}>'.format(self.text)


class Question(BaseQuestion):
    """Represent a question answered by a user other than the logged in user."""

    def return_none_if_unanswered(function):
        def wrapped(self):
            if self.answered:
                return function(self)
        return wrapped

    _answers_xpb = xpb.div.with_class('answers').\
                   p.with_class('answer')

    def __init__(self, question_element):
        super(Question, self).__init__(question_element)
        try:
            self._their_answer_span, self._my_answer_span = \
            self._answers_xpb.span.with_class('text').apply_(self._question_element)
            self._their_note_span, self._my_note_span = \
            self._answers_xpb.span.with_class('note').apply_(self._question_element)
        except:
            pass

    @util.cached_property
    @return_none_if_unanswered
    def their_answer(self):
        return self._their_answer_span.text_content().strip()

    @util.cached_property
    @return_none_if_unanswered
    def my_answer(self):
        return self._my_answer_span.text_content().strip()

    @util.cached_property
    @return_none_if_unanswered
    def their_answer_matches(self):
        return 'not_accepted' not in self._their_answer_span.attrib['class']

    @util.cached_property
    @return_none_if_unanswered
    def my_answer_matches(self):
        return 'not_accepted' not in self._my_answer_span.attrib['class']

    @util.cached_property
    @return_none_if_unanswered
    def their_note(self):
        return self._their_note_span.text_content().strip()

    @util.cached_property
    @return_none_if_unanswered
    def my_note(self):
        return self._my_note_span.text_content().strip()

    _explanation_xpb = xpb.div.span.with_class('note')


class UserQuestion(BaseQuestion):
    """Represent a question answered by the logged in user."""

    _answer_option_xpb = xpb.ul.with_class('self_answers').li
    _explanation_xpb = xpb.div.with_class('your_explanation').p.with_class('value')

    @util.cached_property
    def answer_options(self):
        return [
            AnswerOption(element)
            for element in self._answer_option_xpb.apply_(
                self._question_element
            )
        ]

    @util.cached_property
    def explanation(self):
        try:
            return self._explanation_xpb.get_text_(self._question_element)
        except:
            pass

    @util.cached_property
    def answer_text_to_option(self):
        return {option.text: option for option in self.answer_options}


class AnswerOption(object):

    def __init__(self, option_element):
        self._element = option_element

    @util.cached_property
    def is_users(self):
        return 'mine' in self._element.attrib['class']

    @util.cached_property
    def is_match(self):
        return 'match' in self._element.attrib['class']

    @util.cached_property
    def text(self):
        return self._element.text_content().strip()

    @util.cached_property
    def id(self):
        return self._element.attrib['id'].split('_')[-1]

    def __repr__(self):
        return '<{0}: "{1}" (is_users={2}, is_match={3})>'.format(
            type(self).__name__,
            self.answer_text,
            self.is_users,
            self.is_match
        )


importances = ('not_important', 'little_important', 'somewhat_important',
               'very_important', 'mandatory')


class Questions(object):
    """Interface to accessing and answering questions belonging to the logged
    in user."""

    headers = {
        'accept': 'application/json, text/javascript, */*; q=0.01',
        'accept-encoding': 'gzip,deflate',
        'accept-language': 'en-US,en;q=0.8',
        'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'origin': 'https://www.okcupid.com',
        'referer': 'https://www.okcupid.com/questions',
        'x-requested-with': 'XMLHttpRequest',
    }

    importance_name_to_number = {
        'mandatory': 0,
        'very_important': 1,
        'somewhat_important': 3,
        'little_important': 4,
        'not_important': 5
    }

    path = 'questions/ask'

    def __init__(self, session, importances=importances, user_id=None):
        self.importance_name_to_fetchable = {}
        for importance in importances:
            fetchable = util.Fetchable.fetch_marshall(
                QuestionHTMLFetcher(session, 'questions', **{importance: 1}),
                QuestionProcessor(UserQuestion)
            )
            self.importance_name_to_fetchable[importance] = fetchable
            setattr(self, importance, fetchable)
        self._session = session
        if user_id:
            self._user_id = user_id

    @util.cached_property
    def _user_id(self):
        return self._session.get_current_user_profile().id

    def respond_from_user_question(self, user_question, importance):
        """Respond to a question in exactly the way that is described by
        the given user_question.
        :param user_question: The user question to respond with.
        :type user_question: :class:`.UserQuestion`
        :param importance: The importance that should be used in responding to
                           the question.
        :type importance: int
        """
        user_response_ids = [option.id
                             for option in user_question.answer_options
                             if option.is_users]
        match_response_ids = [option.id
                              for option in user_question.answer_options
                              if option.is_match]
        if len(match_response_ids) == len(user_question.answer_options):
            match_response_ids = 'irrelevant'
        return self.respond(user_question.id, user_response_ids, match_response_ids,
                            importance, note=user_question.explanation or '')

    def respond_from_question(self, question, question_id_to_user_question,
                              importance, switch_acceptable_answer=False):
        user_question = question_id_to_user_question.get(question.id, None)
        if not user_question: return
        option_index = user_question.answer_text_to_option[question.their_answer].id
        self.respond(question.id, [option_index], [option_index], importance)

    def respond(self, question_id, user_response_ids, match_response_ids,
                importance, note='', is_public=1, is_new=1):
        form_data = {
            'ajax': 1,
            'submit': 1,
            'answer_question': 1,
            'skip': 0,
            'show_all': 0,
            'targetid': self._user_id,
            'qid': question_id,
            'answers': user_response_ids,
            'matchanswers': match_response_ids,
            'is_new': is_new,
            'is_public': is_public,
            'note': note,
            'importance': importance,
            'delete_note': 0
        }
        return self._session.okc_post(
            self.path, data=form_data, headers=self.headers
        )

    def clear(self):
        return self._session.okc_post(
            'questions', data={'clear_all': 1}, headers=self.headers
        )


_page_data_xpb = xpb.div.with_class('pages_data')
_current_page_xpb = _page_data_xpb.input(id='questions_pages_page').\
                    select_attribute_('value')
_total_page_xpb = _page_data_xpb.input(id='questions_pages_total').\
                  select_attribute_('value')
_question_xpb = xpb.div.with_class('question')
def QuestionProcessor(question_class):
    return util.PaginationProcessor(question_class, _question_xpb,
                                    _current_page_xpb, _total_page_xpb)


class QuestionHTMLFetcher(object):

    @classmethod
    def from_username(cls, session, username, **kwargs):
        return cls(session, u'profile/{0}/questions'.format(username), **kwargs)

    def __init__(self, session, uri, **additional_parameters):
        self._session = session
        self._uri = uri
        self._additional_parameters = additional_parameters

    def _query_params(self, start_at):
        parameters = {'low': start_at, 'leanmode': 1}
        parameters.update(self._additional_parameters)
        return parameters

    def fetch(self, start_at):
        response = self._session.okc_get(self._uri,
                                         params=self._query_params(start_at))
        return response.content.decode('utf8')


def QuestionFetcher(session, username, question_class=Question,
                    is_user=False, **kwargs):
    if is_user:
        question_class = UserQuestion
    return util.FetchMarshall(
        QuestionHTMLFetcher.from_username(session, username, **kwargs),
        QuestionProcessor(question_class)
    )



import re


sep_replacements = ('\\', '/', '.', '-', ' ', '$', ',', '(', ')')


gentation_to_number = {
    'girls who like guys': '34',
    'guys who like girls': '17',
    'girls who like girls': '40',
    'guys who like guys': '20',
    'both who like bi guys': '54',
    'both who like bi girls': '57',
    'straight girls only': '2',
    'straight guys only': '1',
    'gay girls only': '8',
    'gay guys only': '4',
    'bi girls only': '32',
    'bi guys only': '16',
    'bi guys and girls': '48',
    'everybody': '63',
    '': '63'
}


looking_for_re_numbers = ((re.compile("[Ff]riends"), 1),
                          (re.compile("[Ll]ong.*[Dd]ating"), 2),
                          (re.compile("[Ss]hort.*[Dd]ating"), 3),
                          (re.compile("[Ss]ex"), 6))


has_kids = {
    "has a kid": {"addition": 33686018, "power": 1},
    "has kids": {"addition": 67372036, "power": 2},
    "doesn't have kids": {"addition": 1077952576, "power": 6},
}


wants_kids = {
    "might want kids": {"addition": 18176, "power": 8},
    "wants kids": {"addition": 4653056, "power": 16},
    "doesn't want kids": {"addition": 1191182384, "power": 24},
}


language_map = {
    'Afrikaans': '2',
    'Albanian': '3',
    'Ancient Greek': '27',
    'Arabic': '4',
    'Armenian': '76',
    'Basque': '5',
    'Belarusan': '6',
    'Bengali': '7',
    'Breton': '8',
    'Bulgarian': '9',
    'C++': '71',
    'Catalan': '10',
    'Cebuano': '11',
    'Chechen': '12',
    'Chinese': '13',
    'Croatian': '14',
    'Czech': '15',
    'Danish': '16',
    'Dutch': '17',
    'English': '74',
    'Esperanto': '18',
    'Estonian': '19',
    'Farsi': '20',
    'Finnish': '21',
    'French': '22',
    'Frisian': '23',
    'Georgian': '24',
    'German': '25',
    'Greek': '26',
    'Gujarati': '77',
    'Hawaiian': '28',
    'Hebrew': '29',
    'Hindi': '75',
    'Hungarian': '30',
    'Icelandic': '31',
    'Ilongo': '32',
    'Indonesian': '33',
    'Irish': '34',
    'Italian': '35',
    'Japanese': '36',
    'Khmer': '37',
    'Korean': '38',
    'LISP': '72',
    'Latin': '39',
    'Latvian': '40',
    'Lithuanian': '41',
    'Malay': '42',
    'Maori': '43',
    'Mongolian': '44',
    'Norwegian': '45',
    'Occitan': '46',
    'Other': '73',
    'Persian': '47',
    'Polish': '48',
    'Portuguese': '49',
    'Romanian': '50',
    'Rotuman': '51',
    'Russian': '52',
    'Sanskrit': '53',
    'Sardinian': '54',
    'Serbian': '55',
    'Sign Language': '1',
    'Slovak': '56',
    'Slovenian': '57',
    'Spanish': '58',
    'Swahili': '59',
    'Swedish': '60',
    'Tagalog': '61',
    'Tamil': '62',
    'Thai': '63',
    'Tibetan': '64',
    'Turkish': '65',
    'Ukrainian': '66',
    'Urdu': '67',
    'Vietnamese': '68',
    'Welsh': '69',
    'Yiddish': '70',
}


language_map_2 = {
    'afrikaans': 'af',
    'albanian': 'sq',
    'arabic': 'ar',
    'armenian': 'hy',
    'basque': 'eu',
    'belarusan': 'be',
    'bengali': 'bn',
    'breton': 'br',
    'bulgarian': 'bg',
    'catalan': 'ca',
    'cebuano': '11',
    'chechen': 'ce',
    'chinese': 'zh',
    'c++': '71',
    'croatian': 'hr',
    'czech': 'cs',
    'danish': 'da',
    'dutch': 'nl',
    'english': 'en',
    'esperanto': 'eo',
    'estonian': 'et',
    'farsi': '20',
    'finnish': 'fi',
    'french': 'fr',
    'frisian': '23',
    'georgian': '24',
    'german': 'de',
    'greek': 'el',
    'gujarati': 'gu',
    'ancient greek': '27',
    'hawaiian': '28',
    'hebrew': 'he',
    'hindi': 'hi',
    'hungarian': 'hu',
    'icelandic': 'is',
    'ilongo': '32',
    'indonesian': 'id',
    'irish': 'ga',
    'italian': 'it',
    'japanese': 'ja',
    'khmer': '37',
    'korean': 'ko',
    'latin': 'la',
    'latvian': 'lv',
    'lisp': '72',
    'lithuanian': 'lt',
    'malay': 'ms',
    'maori': 'mi',
    'mongolian': 'mn',
    'norwegian': 'no',
    'occitan': 'oc',
    'other': '73',
    'persian': 'fa',
    'polish': 'pl',
    'portuguese': 'pt',
    'romanian': 'ro',
    'rotuman': '51',
    'russian': 'ru',
    'sanskrit': 'sa',
    'sardinian': '54',
    'serbian': 'sr',
    'sign language': '1',
    'slovak': 'sk',
    'slovenian': 'sl',
    'spanish': 'es',
    'swahili': 'sw',
    'swedish': 'sv',
    'tagalog': 'tl',
    'tamil': 'ta',
    'thai': 'th',
    'tibetan': 'bo',
    'turkish': 'tr',
    'ukrainian': 'uk',
    'urdu': 'ur',
    'vietnamese': 'vi',
    'welsh': 'cy',
    'yiddish': 'yi'
}


binary_lists = {
    'smokes': ['11', 'yes', 'sometimes', 'when drinking', 'trying to quit', 'no'],
    'drinks': ['12', 'very often', 'often', 'socially', 'rarely', 'desperately',
               'not at all'],
    'drugs': ['13', 'never', 'sometimes', 'often'],
    'education': ['19', 'high school', 'two-year college', 'college/university',
                  'masters program', 'law school', 'medical school',
                  'ph.d program'],
    'job': ['15', 'student', 'art / music / writing', 'banking / finance',
            'administration', 'technology', 'construction', 'education',
            'entertainment / media', 'management', 'hospitality', 'law',
            'medicine', 'military', 'politics / government',
            'sales / marketing', 'science / engineering', 'transporation',
            'retired'],
    'income': ['14', 'less than $20,000', '$20,000-$30,000', '$30,000-$40,000',
               '$40,000-$50,000', '$50,000-$60,000', '$60,000-$70,000',
               '$70,000-$80,000', '$80,000-$100,000', '$100,000-$150,000',
               '$150,000-$250,000', '$250,000-$500,000', '$500,000-$1,000,000',
               'More than $1,000,000'],
    'religion': ['8', 'agnostic', 'atheist', 'christian', 'jewish', 'catholic',
                 'muslim', 'hindu', 'buddhist', 'other'],
    'monogamy': ['73', 'monogamous', 'non-monogamous'],
    'diet': ['54', 'vegetarian', 'vegan', 'kosher', 'halal'],
    'sign': ['21', 'acquarius', 'pisces', 'aries', 'taurus', 'gemini',
             'cancer', 'leo', 'virgo', 'libra', 'scorpio', 'sagittarius',
             'capricorn'],
    'ethnicity': ['9', 'asian', 'middle eastern', 'black', 'native american',
                  'indian', 'pacific islander', 'hispanic/latin', 'white',
                  'human (other)']
    }


dogs = ['owns dogs', 'likes dogs', 'dislikes dogs']
cats = ['owns cats', 'likes cats', 'dislikes cats']


def get_height_query(height_min, height_max):
    '''Convert from inches to millimeters/10'''
    min_int = 0
    max_int = 99999
    if height_min is not None:
        min_int = int(height_min) * 254
    if height_max is not None:
        max_int = int(height_max) * 254
    return '10,{0},{1}'.format(str(min_int), str(max_int))


def get_options_query(type, inputs):
    for char in sep_replacements:
        inputs = [input.replace(char, '') for input in inputs]
    inputs = [input.lower() for input in inputs]
    if type in binary_lists:
        start = 1
        option_list = binary_lists[type]
    if type == 'drugs':
        start = 0
    if type == 'diet':
        start = 2
    int_to_return = 0
    for power, option in enumerate(option_list[1:], start=start):
        for char in sep_replacements:
            option = option.replace(char, '')
        if option in inputs:
            int_to_return += 2**power
    return '{0},{1}'.format(option_list[0], int_to_return)


def get_pet_queries(pets):
    for char in sep_replacements:
        pets = [pet.replace(char, '') for pet in pets]
    pets = [pet.lower() for pet in pets]
    dog_int = 0
    cat_int = 0
    for power, option in enumerate(dogs, start=1):
        for char in sep_replacements:
            option = option.replace(char, '')
        if option in pets:
            dog_int += 2**power
    for power, option in enumerate(cats, start=1):
        for char in sep_replacements:
            option = option.replace(char, '')
        if option in pets:
            cat_int += 2**power
    dog_query = '16,{0}'.format(dog_int)
    cat_query = '17,{0}'.format(cat_int)
    return dog_query, cat_query


def get_kids_query(kids):
    kids = [kid.lower() for kid in kids]
    kid_int = 0
    include_has = False
    include_wants = False
    for option in kids:
        if option in has_kids:
            include_has = True
        elif option in wants_kids:
            include_wants = True
    if include_has and not include_wants:
        for option in kids:
            if option in has_kids:
                kid_int += has_kids[option]['addition']
    elif include_wants and not include_has:
        for option in kids:
            if option in wants_kids:
                kid_int += wants_kids[option]['addition']
    elif include_wants and include_has:
        for has_option in kids:
            if has_option in has_kids:
                for wants_option in kids:
                    if wants_option in wants_kids:
                        kid_int += 2**(has_kids[has_option]['power'] + wants_kids[wants_option]['power'])
    return '18,{0}'.format(kid_int)


def get_language_query(language):
    return '22,{0}'.format(language_map[language.title()])


hour = 60*60
join_date_string_to_int = {
    'hour': hour,
    'day': hour*24,
    'week': hour*24*7,
    'month': hour*24*30,
    'year': 365*24*hour
}


def get_join_date_query(date):
    date_int = 0
    if isinstance(date, str) and not date.isdigit():
        if date in join_date_string_to_int:
            date_int = join_date_string_to_int[date]
        else:
            date_int = int(date)
    else:
        date_int = date
    return '6,{0}'.format(date_int)



import argparse
import pkg_resources
import sys

from invoke import cli as invoke
import IPython

from . import tasks
from . import util
from .attractiveness_finder import AttractivenessFinder
from .photo import PhotoUploader
from .session import Session
from .statistics import Statistics
from .user import User
from .util import save_file


def interactive():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--version', action='store_true',
                        help='Display the version number of okcupyd')
    util.add_command_line_options(parser.add_argument)
    args = parser.parse_args()
    if args.version:
        print(pkg_resources.get_distribution('okcupyd').version)
        return sys.exit()
    util.handle_command_line_options(args)
    util.get_credentials()
    u = User()
    IPython.embed()
    # try:
    #     args, collection, parser_contexts = invoke.parse(sys.argv, collection=tasks.ns)
    # except invoke.Exit as e:
    #     # 'return' here is mostly a concession to testing. Meh :(
    #     # TODO: probably restructure things better so we don't need this?
    #     return sys.exit(e.code)
    # executor = invoke.Executor(tasks.ns, invoke.Context(**invoke.derive_opts(args)))

    # _tasks = invoke.tasks_from_contexts(parser_contexts, tasks.ns)
    # dedupe = not args['no-dedupe'].value
    # executor.execute(*_tasks, dedupe=dedupe)


__all__ = ('User', 'AttractivenessFinder', 'Statistics',
           'save_file', 'interactive', 'PhotoUploader', 'Session')



import copy
import logging
import os
import zlib

from six.moves import urllib
from wrapt import decorator
import simplejson
import vcr

from okcupyd import settings
from okcupyd import util

log = logging.getLogger(__name__)

TESTING_USERNAME = 'username'
TESTING_PASSWORD = 'password'
WBITS = 16 + zlib.MAX_WBITS


SHOULD_SCRUB = False
REPLACEMENTS = []
REMOVE_OLD_CASSETTES = False


@decorator
def check_should_scrub(function, instance, args, kwargs):
    if SHOULD_SCRUB:
        return function(*args)
    else:
        return args[0] # The request or response


@util.curry
def remove_headers(request, headers_to_remove=()):
    headers = copy.copy(request.headers)
    headers_to_remove = [h.lower() for h in headers_to_remove]
    keys = [k for k in headers if k.lower() in headers_to_remove]
    if keys:
        for k in keys:
            headers.pop(k)
        request.headers = headers
    return request


def scrub_request_body(request):
    if urllib.parse.urlsplit(request.uri).path == '/login':
        request.body = scrub_query_string(request.body)
    request.uri = scrub_uri(request.uri)
    return request


def scrub_uri(uri):
    replaced = util.replace_all_case_insensitive(uri, settings.USERNAME,
                                                 TESTING_USERNAME)
    return util.replace_all_case_insensitive(replaced, settings.PASSWORD,
                                             TESTING_PASSWORD)


def scrub_query_string(query_string):
    request_dict = urllib.parse.parse_qs(query_string)
    if 'password' not in request_dict:
        return query_string

    for key in request_dict:
        request_dict[key] = request_dict[key][0]
    request_dict['username'] = TESTING_USERNAME
    request_dict['password'] = TESTING_PASSWORD
    return urllib.parse.urlencode(request_dict)


def gzip_string(incoming):
    if isinstance(incoming, str) and bytes is not str:
        incoming = bytes(incoming, 'utf8')
    else:
        incoming = incoming.encode('utf8')
    compress_object = zlib.compressobj(6, zlib.DEFLATED, WBITS)
    start = compress_object.compress(incoming)
    end = compress_object.flush()
    if not isinstance(start, str):
        return start + end
    return ''.join([start, end])


def scrub_response_headers(response):
    for item in ('location', 'Location'):
        if item in response['headers']:
            response['headers'][item] = [scrub_uri(uri)
                                         for uri in response['headers'][item]]
    return response


def replace_json_fields(body):
    try:
        response_dict = simplejson.loads(body)
    except:
        return body

    if 'screenname' not in response_dict:
        return body
    if response_dict['screenname'] is not None:
        response_dict['screenname'] = TESTING_USERNAME
    response_dict['userid'] = 1
    response_dict['thumbnail'] = ''
    return simplejson.dumps(response_dict)


def scrub_response(response):
    if not SHOULD_SCRUB:
        return response
    response = response.copy()
    response = scrub_response_headers(response)

    body = response['body']['string']
    try:
        body = zlib.decompress(response['body']['string'], WBITS).decode('utf8')
    except:
        should_recompress = False
    else:
        should_recompress = True

    body = replace_json_fields(body)
    body = util.replace_all_case_insensitive(body, settings.USERNAME,
                                             TESTING_USERNAME)

    if should_recompress:
        body = gzip_string(body)
    response['body']['string'] = body
    return response


before_record = check_should_scrub(util.compose(
    scrub_request_body, remove_headers(headers_to_remove=('Set-Cookie', 'Cookie'))
))


def _match_search_query(left, right):
    left_filter = set([value for param_name, value in left
                       if 'filter' in param_name])
    right_filter = set([value for param_name, value in right
                        if 'filter' in param_name])
    left_rest = set([(param_name, value) for param_name, value in left
                     if 'filter' not in param_name])
    right_rest = set([(param_name, value) for param_name, value in right
                      if 'filter' not in param_name])

    log.info(simplejson.dumps(
        {
            'filter_differences': list(left_filter.symmetric_difference(right_filter)),
            'rest_differences': list(left_rest.symmetric_difference(right_rest)),
        }
    ))

    return left_filter == right_filter and left_rest == right_rest


def match_search_query(left, right):
    return _match_search_query(left.query, right.query)


def body_as_query_string(left, right):
    try:
        left_qs_items = list(urllib.parse.parse_qs(left.body).items())
        right_qs_items = list(urllib.parse.parse_qs(right.body).items())
    except Exception as exc:
        log.debug(exc)
        return left.body == right.body
    else:
        left_qs_items = [(k, tuple(v)) for k, v in left_qs_items]
        right_qs_items = [(k, tuple(v)) for k, v in right_qs_items]
        return _match_search_query(left_qs_items, right_qs_items)


okcupyd_vcr = vcr.VCR(match_on=('path', 'method', 'match_search_query',
                                'body_as_query_string'),
                      before_record=(before_record,),
                      before_record_response=scrub_response)
okcupyd_vcr.register_matcher('body_as_query_string', body_as_query_string)
okcupyd_vcr.register_matcher('match_search_query', match_search_query)
match_on_no_body = list(filter(lambda x: 'body' not in x, okcupyd_vcr.match_on))


class cassette(object):

    base_path = os.path.dirname(__file__)

    @classmethod
    def path(cls, cassette_name):
        return os.path.join(cls.base_path,
                            'vcr_cassettes', '{0}.yaml'.format(cassette_name))


@util.curry
def use_cassette(function=None, cassette_name=None, *args, **kwargs):
    if cassette_name is None:
        assert function, 'Must supply function if no cassette name given'
        cassette_name = function.__name__
    path = cassette.path(cassette_name)
    context_decorator = okcupyd_vcr.use_cassette(path, *args, **kwargs)
    if function:
        return context_decorator(function)
    return context_decorator



from sqlalchemy import and_, or_
from sqlalchemy.orm import aliased

from okcupyd.db import model, with_txn


def have_messaged_by_username_no_txn(session, username_one, username_two):
    user_one = aliased(model.User)
    user_two = aliased(model.User)
    return session.query(model.MessageThread).\
        join(user_one, user_one.id == model.MessageThread.initiator_id).\
        join(user_two, user_two.id == model.MessageThread.respondent_id).\
        filter(or_(
            and_(user_one.handle == username_one, user_two.handle == username_two),
            and_(user_one.handle == username_two, user_two.handle == username_one),
        )).exists()


@with_txn
def have_messaged_by_username(session, username_one, username_two):
    return session.query(
        have_messaged_by_username_no_txn(session, username_one, username_two)
    ).scalar()


import logging
import re

from lxml import html
import simplejson

from . import util
from . import helpers
from .session import Session
from .xpath import xpb


log = logging.getLogger(__name__)


@util.curry
class PhotoUploader(object):

    _uri = 'ajaxuploader'

    def __init__(self, session=None, user_id=None, authcode=None):
        self._session = session or Session.login()
        if user_id:
            self._user_id = user_id
        if authcode:
            self._authcode = authcode

    @util.cached_property
    def _photo_tree(self):
        return html.fromstring(self._session.okc_get(
            u'profile/{0}/photos#upload'.format(self._session.log_in_name)
        ).content)

    @util.cached_property
    def _authcode(self):
        return helpers.get_authcode(self._photo_tree)

    @util.cached_property
    def _user_id(self):
        return helpers.get_current_user_id(self._photo_tree)

    @property
    def pic_id(self):
        return self._response_dict['id']

    @property
    def height(self):
        return self._response_dict['height']

    @property
    def width(self):
        return self._response_dict['width']

    def upload(self, incoming):
        if isinstance(incoming, Info):
            return self.upload_from_info(incoming)
        if hasattr(incoming, 'read'):
            return self.upload_file(incoming)
        return self.upload_by_filename(incoming)

    def upload_from_info(self, info):
        response = self._session.get(info.jpg_uri, stream=True)
        self.upload_file(response.raw)

    def upload_by_filename(self, filename):
        with open(filename, 'rb') as file_object:
            self.upload(file_object)

    def upload_file(self, file_object):
        files = {'file': ('my_photo.jpg', file_object,
                          'image/jpeg', {'Expires': '0'})}
        response = self._session.okc_post(self._uri, files=files)
        response_script_text = xpb.script.get_text_(
            html.fromstring(response.content)
        )
        self._response_dict = self._get_response_json(response_script_text)
        return self._response_dict

    def _get_response_json(self, response_text):
        start = response_text.find('res =') + 5
        end = response_text.find('};') + 1
        response_text = response_text[start:end]
        log.info(simplejson.dumps({'photo_upload_response': response_text}))
        return simplejson.loads(response_text)

    def _confirm_parameters(self, thumb_nail_left=None, thumb_nail_top=None,
                            thumb_nail_right=None, thumb_nail_bottom=None):
        return {
            'userid': self._user_id,
            'albumid': 0,
            'authcode': self._authcode,
            'okc_api': 1,
            'picid': self.pic_id,
            'picture.add_ajax': 1,
            'use_new_upload': 1,
            'caption': '',
            'height': self.height,
            'width': self.width,
            'tn_upper_left_x': thumb_nail_left or 0,
            'tn_upper_left_y': thumb_nail_top or 0,
            'tn_lower_right_x': thumb_nail_right or self.width,
            'tn_lower_right_y': thumb_nail_bottom or self.height
        }

    def confirm(self, **kwargs):
        return self._session.okc_get('photoupload',
                                     params=self._confirm_parameters(**kwargs))

    def upload_and_confirm(self, incoming, **kwargs):
        self.upload(incoming)
        if isinstance(incoming, Info):
            kwargs.setdefault('thumb_nail_left', incoming.thumb_nail_left)
            kwargs.setdefault('thumb_nail_top', incoming.thumb_nail_top)
            kwargs.setdefault('thumb_nail_right', incoming.thumb_nail_right)
            kwargs.setdefault('thumb_nail_bottom', incoming.thumb_nail_bottom)
        self.confirm(**kwargs)
        return self._response_dict


class Info(object):

    base_uri = "http://k0.okccdn.com/php/load_okc_image.php/images/"

    cdn_re = re.compile("http://.*okccdn.com.*images/"
                        "[0-9]*x[0-9]*/[0-9]*x[0-9]*/"
                        "(?P<tnl>[0-9]*?)x(?P<tnt>[0-9]*?)/"
                        "(?P<tnr>[0-9]*)x(?P<tnb>[0-9]*)/0/"
                        "(?P<id>[0-9]*).webp\?v=2")

    @classmethod
    def from_cdn_uri(cls, cdn_uri):
        match = cls.cdn_re.match(cdn_uri)
        return cls(match.group('id'), match.group('tnl'), match.group('tnt'),
            match.group('tnr'), match.group('tnb'))

    def __init__(self, photo_id, tnl, tnt, tnr, tnb):
        self.id = photo_id
        self.thumb_nail_left = int(tnl)
        self.thumb_nail_top = int(tnt)
        self.thumb_nail_right = int(tnr)
        self.thumb_nail_bottom = int(tnb)

    @property
    def jpg_uri(self):
        return '{0}{1}.jpg'.format(self.base_uri, self.id)

    def __repr__(self):
        return 'photo.{0}({1}, {2}, {3}, {4}, {5})'.format(
            type(self).__name__,
            self.id,
            self.thumb_nail_left,
            self.thumb_nail_top,
            self.thumb_nail_right,
            self.thumb_nail_bottom
        )



from . import helpers
from . import util
from .attractiveness_finder import AttractivenessFinder
from .messaging import ThreadFetcher
from .photo import PhotoUploader
from .profile import Profile
from .question import Questions
from .search import SearchFetchable, search
from .session import Session
from .xpath import xpb


class User(object):
    """Encapsulate a logged in okcupid user."""

    @classmethod
    def from_credentials(cls, username, password):
        """
        :param username: The username to log in with.
        :type username: str
        :param password: The password to log in with.
        :type password: str
        """
        return cls(Session.login(username, password))

    _visitors_xpb = xpb.div.with_class('user_info').\
                   div.with_class('profile_info').div.with_class('username').\
                   a.with_class('name').text_

    _visitors_current_page_xpb = xpb.div.with_class('pages').\
                                 span.with_class('curpage').text_
    _visitors_total_page_xpb = xpb.div.with_class('pages').\
                               a.with_class('last').text_

    def __init__(self, session=None):
        """
        :param session: A logged in :class:`okcupyd.session.Session`
        """
        self._session = session or Session.login()
        self._message_sender = helpers.Messager(self._session)
        assert self._session.log_in_name is not None, (
            "The session provided to the user constructor must be logged in."
        )
        #: A :class:`okcupyd.profile.Profile` belonging to the logged in user.
        self.profile = Profile(self._session, self._session.log_in_name)

        #: A :class:`okcupyd.util.fetchable.Fetchable` of
        #: :class:`okcupyd.messaging.MessageThread` objects corresponding to
        #: messages that are currently in the user's inbox.
        self.inbox = util.Fetchable(ThreadFetcher(self._session, 1))
        #: A :class:`okcupyd.util.fetchable.Fetchable` of
        #: :class:`okcupyd.messaging.MessageThread` objects corresponding to
        #: messages that are currently in the user's outbox.
        self.outbox = util.Fetchable(ThreadFetcher(self._session, 2))
        #: A :class:`okcupyd.util.fetchable.Fetchable` of
        #: :class:`okcupyd.messaging.MessageThread` objects corresponding to
        #: messages that are currently in the user's drafts folder.
        self.drafts = util.Fetchable(ThreadFetcher(self._session, 4))

        #: A :class:`okcupyd.util.fetchable.Fetchable` of
        #: :class:`okcupyd.profile.Profile` objects of okcupid.com users that
        #: have visited this user's profile.
        self.visitors = util.Fetchable.fetch_marshall(
            util.GETFetcher(self._session, 'visitors',
                            lambda start_at: {'low': start_at}),
            util.PaginationProcessor(
                lambda user: Profile(self._session, user), self._visitors_xpb,
                self._visitors_current_page_xpb, self._visitors_total_page_xpb,
            )
        )

        #: A :class:`okcupyd.question.Questions` that is instantiated with
        #: this instance's session.
        self.questions = Questions(self._session)

        #: A :class:`okcupyd.attractiveness_finder.AttractivenessFinder`
        #: that is instantiated with this instance's session.
        self.attractiveness_finder = AttractivenessFinder(self._session)

        #: A :class:`okcupyd.photo.PhotoUploader` that is instantiated with
        #: this instance's session.
        self.photo = PhotoUploader(self._session)

    def get_profile(self, username):
        """Get the :class:`okcupyd.profile.Profile` associated with the supplied
        username.
        :param username: The username of the profile to retrieve.
        """
        return self._session.get_profile(username)

    @property
    def username(self):
        """Return the username associated with the :class:`.User`."""
        return self.profile.username

    def message(self, username, message_text):
        """Message an okcupid user. If an existing conversation between the
        logged in user and the target user can be found, reply to that thread
        instead of starting a new one.

        :param username: The username of the user to which the message should
                         be sent.
        :param message_text: The body of the message.

        """
        # Try to reply to an existing thread.
        if not isinstance(username, str):
            username = username.username
        for thread in sorted(set(self.inbox + self.outbox),
                             key=lambda t: t.datetime, reverse=True):
            if thread.correspondent.lower() == username.lower():
                thread.reply(message_text)
                return

        return self._message_sender.send(username, message_text)

    def search(self, **kwargs):
        """Return a :class:`.util.fetchable.Fetchable` that wraps a
        :function:`okcupyd.search.SearchFetcher`.

        Defaults for `gender`, `gentation`, `location` and `radius` will
        be provided to the constructor of the
        :class:`okcupyd.search.SearchManager` unless they are explicitly
        provided.

        :param kwargs: Arguments that should be passed to the constructor of
                       :class:`okcupyd.search.SearchManager`

        :param kwargs: arguments to pass to :meth:`search_manager`.
        """
        kwargs.setdefault('gender', self.profile.gender[0])
        gentation = helpers.get_default_gentation(self.profile.gender,
                                              self.profile.orientation)
        kwargs.setdefault('gentation', gentation)
        kwargs.setdefault('location', self.profile.location)
        kwargs.setdefault('radius', 25)
        if 'count' in kwargs:
            count = kwargs.pop('count')
            return search(session=self._session, count=count, **kwargs)
        return SearchFetchable(self._session, **kwargs)

    def quickmatch(self):
        """Return a :class:`okcupyd.profile.Profile` obtained by visiting the
        quickmatch page.
        """
        response = self._session.okc_get('quickmatch', params={'okc_api': 1})
        return Profile(self._session, response.json()['sn'])

    def __repr__(self):
        return 'User("{0}")'.format(self.profile.username)



from .search import search
from .session import Session
from . import settings
from . import util


class _AttractivenessFinder(object):

    def __init__(self, session=None):
        self._session = session or Session.login(settings.AF_USERNAME,
                                                 settings.AF_PASSWORD)

    def find_attractiveness(self, username, accuracy=1000,
                            _lower=0, _higher=10000):
        average = (_higher + _lower)//2
        if _higher - _lower <= accuracy:
            return average

        results = search(self._session,
                         count=1,
                         gentation='everybody',
                         keywords=username,
                         attractiveness_min=average,
                         attractiveness_max=_higher,)

        if results:
            return self.find_attractiveness(username, accuracy,
                                            average, _higher)
        else:
            return self.find_attractiveness(username, accuracy,
                                            _lower, average)

    __call__ = find_attractiveness


class AttractivenessFinderDecorator(object):

    def __init__(self, attractiveness_finder=None):
        self._finder = attractiveness_finder or _AttractivenessFinder()

    def __getattr__(self, attr):
        return getattr(self._finder, attr)

    def __call__(self, *args, **kwargs):
        return self.find_attractiveness(*args, **kwargs)


class CheckForExistenceAttractivenessFinder(AttractivenessFinderDecorator):

    def _check_for_existence(self, username):
        return bool(search(self._session,
                           gentation='everybody',
                           keywords=username,
                           count=1))


    def find_attractiveness(self, username, *args, **kwargs):
        if self._check_for_existence(username):
            return self._finder(username, *args, **kwargs)


class RoundedAttractivenessFinder(AttractivenessFinderDecorator):

    def find_attractiveness(self, *args, **kwargs):
        unrounded = self._finder.find_attractiveness(*args, **kwargs)
        if unrounded is not None:
            return int(round(float(unrounded)/1000, 0)*1000)


class CachedAttractivenessFinder(AttractivenessFinderDecorator):

    def __init__(self, attractiveness_finder=None):
        self._finder = attractiveness_finder or _AttractivenessFinder()
        self._cache = {}

    def find_attractiveness(self, username, **kwargs):
        if username not in self._cache:
            self._cache[username] = self._finder(username, **kwargs)
        return self._cache[username]


AttractivenessFinder = util.compose(CachedAttractivenessFinder,
                                    RoundedAttractivenessFinder,
                                    CheckForExistenceAttractivenessFinder,
                                    _AttractivenessFinder)
